# Copyright (C) 2013 Duncan Macleod

# GLUE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
# 
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

"""The glue.auth module provides methods to communicate with data
stored behind the LIGO.ORG authentication system
"""

import os
import sys
import stat
from six.moves import urllib
import six.moves.http_cookiejar

from .saml import HTTPNegotiateAuthHandler
from .. import git_version

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__credits__ = "Scott Koranda <scott.koranda@ligo.org>"
__date__ = git_version.date
__version__ = git_version.id

COOKIE_JAR = '/tmp/%s_cookies' % os.getenv('USER')
LIGO_LOGIN_URL = 'login.ligo.org'


def request_ligodotorg(url, debug=False):
    """Request the given URL using LIGO.ORG SAML authentication.

    This requires an active Kerberos ticket for the user, to get one:

    $ kinit albert.einstein@LIGO.ORG

    Parameters
    ----------
    url : `str`
        URL path for request
    debug : `bool`, optional
        Query in verbose debuggin mode, default `False`

    Returns
    -------
    urllib.addinfourl
        file object containing output data, use .read() to extract
        text content
    """
    # set debug to 1 to see all HTTP(s) traffic
    debug = int(debug)

    # need an instance of HTTPS handler to do HTTPS
    httpsHandler = HTTPSHandler(debuglevel = debug)

    # use a cookie jar to store session cookies
    jar = six.moves.http_cookiejar.LWPCookieJar()

    # if a cookier jar exists open it and read the cookies
    # and make sure it has the right permissions
    if os.path.exists(COOKIE_JAR):
        os.chmod(COOKIE_JAR, stat.S_IRUSR | stat.S_IWUSR)

        # set ignore_discard so that session cookies are preserved
        jar.load(COOKIE_JAR, ignore_discard = True)

    # create a cookie handler from the cookier jar
    cookie_handler = urllib.request.HTTPCookieProcessor(jar)
    # need a redirect handler to follow redirects
    redirectHandler = urllib.request.HTTPRedirectHandler()

    # need an auth handler that can do negotiation.
    # input parameter is the Kerberos service principal.
    auth_handler = HTTPNegotiateAuthHandler(service_principal='HTTP@%s'
                                                            % (LIGO_LOGIN_URL))

    # create the opener.
    opener = urllib.request.build_opener(auth_handler, cookie_handler, httpsHandler,
                                  redirectHandler)

    # prepare the request object
    request = urllib.request.Request(url)

    # use the opener and the request object to make the request.
    response = opener.open(request)

    # save the session cookies to a file so that they can
    # be used again without having to authenticate
    jar.save(COOKIE_JAR, ignore_discard=True)

    return response




#
# lightweight database dumper dmt segment client
#
#
# Copyright (C) 2005 Duncan Brown
# 
# This is part of the Grid LSC User Environment (GLUE)
# 
# GLUE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
# 
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function
import sys
import os
import getopt
import re

try:
  from glue.segmentdb import segmentdb_utils
except ImportError as e:
  print("""
Error: unable to import modules from glue.

Check that glue is correctly installed and in your PYTHONPATH.

%s
""" % e, file=sys.stderr)
  sys.exit(1)


def usage():
  msg = """\
Usage: dmtdq_seg_insert [OPTIONS]

  -t, --segment-url       PROTOCOL:HOST Supported protocols include: https and ldbd. For example, "--segment-url=https://segdb.ligo.caltech.edu" or "--segment-url=ldbd://segdb.ligo.caltech.edu"
  -f, --file              file to insert (deleted upon success)
  -p, --ping              ping the server
  -h, --help              print this message

\
"""
  print(msg)

def help():
  usage()
  sys.exit(0)

shortop = "t:f:ph"
longop = [
  "segment-url=",
  "file=",
  "ping",
  "help"
  ]

try:
  opts, args = getopt.getopt(sys.argv[1:], shortop, longop)
except getopt.GetoptError as e:
  print("Error: %s \n" % str(e), file=sys.stderr)
  usage()
  sys.exit(1)

# defaults
segment_url = None
infile = None
ping = None

# environment variables override defaults but not
# command line options
try:
  segment_url = os.environ['LDBD_SERVER']
except:
  pass

for o, a in opts:
  if o in ("-h", "--help"):
    help()
  elif o in ("-t", "--segment-url"):
    segment_url = a
  elif o in ("-f", "--file"):
    infile = a
  elif o in ("-p", "--ping"):
    ping = 1

# determine server and port
if not segment_url:
  print("No --segment-url specified", file=sys.stderr)
  print("Enter 'dmtdq_seg_insert --help' for usage", file=sys.stderr)
  sys.exit(1)


if not ping and not infile:
  print("\nNo file specified\n")
  usage()
  sys.exit(1)

# open connection to LDBD(W)Server
myClient = segmentdb_utils.setup_database(segment_url)

try:
  if ping:
    # ping the server and print the response
    print(myClient.ping())

  else:
    # open the input file and send it to the server
    fh = open(infile,'r')
    xmltext = fh.read()
    myClient.insertdmt(xmltext)
    fh.close()

except Exception as e:
  print("Error inserting segments: %s" % str(e), file=sys.stderr)
  sys.exit(1)

if infile:
  os.unlink(infile)

sys.exit(0)



"""
The LDBDWServer module provides an API for responding to request from the LDBDWClient by connecting to the DB2 database.

This file is part of the Grid LSC User Environment (GLUE)

GLUE is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""
from glue import git_version
__date__ = git_version.date
__version__ = git_version.id

import os
import sys
import re
import types
try:
    import pyRXP
except ImportError:
    import pyRXPU as pyRXP
import socket
import six.moves.cPickle
import logging
import logging.handlers
import six.moves.configparser
import M2Crypto
import selector
import simplejson

try:
    from cjson import (decode, encode)
except ImportError:
    from json import (loads as decode, dumps as encode)

from glue import ldbd

# default configuration
configuration = {
  'gridmap' : '/etc/ldbd/grid-mapfile',
  'gridmap_insert' : '/etc/ldbd/grid-mapfile',
  'gridmap_insertmap' : '/etc/ldbd/grid-mapfile',
  'gridmap_insertdmt' : '/etc/ldbd/grid-mapfile',
  'dbname' : 'ldbd_tst',
  'dbuser' : 'grid',
  'dbpasswd' : '',
  'logfile' : '/var/log/ldbd/ldbdd.log',
  'logmaxbytes' : 1024 * 1024 * 1,
  'max_client_byte_string' : 1048576,
  'logbackupcount' : 5,
  'loglevel' : 'DEBUG'
  }

# grab configuration from file
myConfigParser = six.moves.configparser.ConfigParser()

try:
    ldbdserver_ini = os.environ.get('LDBD_CONFIG_DIR', '/usr3/ldbd/etc')
    config_file = os.path.join(ldbdserver_ini, "ldbdserver.ini")
    myConfigParser.read(config_file)
except Exception as e:
    sys.stder.write("Error: unable to read configuration file : %s\n" % config_file)
    sys.exit(1)

for k in configuration.keys():
    try:
        value = myConfigParser.get('ldbdd',k)
    except six.moves.configparser.NoOptionError as e:
        sys.stderr.write("Error: missing configuration option %s: %s\n" % (k, e))
        sys.exit(1)
    try:
        configuration[k] = eval(value)
    except Exception as e:
        configuration[k] = value

# set up logging
logFilePath = configuration['logfile']
logMaxBytes = configuration['logmaxbytes']
logBackupCount = configuration['logbackupcount']
logLevel = configuration['loglevel']

# set up logging for the root
rootlogger = logging.getLogger('LDBD')

handler = logging.handlers.RotatingFileHandler(logFilePath, 'a', logMaxBytes, logBackupCount)
formatter = logging.Formatter('%(asctime)s %(name)-12s %(levelname)-8s %(message)s')
handler.setFormatter(formatter)

rootlogger.addHandler(handler)
rootlogger.setLevel(eval("logging.%s" % logLevel))

# create logger to use with correct prefix
logger = logging.getLogger('LDBD.Server')
logger.info("LDBD server started")

# initialize the database hash table
dbobj = ldbd.LIGOMetadataDatabase(configuration['dbname'])
max_bytes = configuration['max_client_byte_string']

# create the xml parser
xmlparser = pyRXP.Parser()

def dtd_uri_callback(uri):
  if uri == 'http://ldas-sw.ligo.caltech.edu/doc/ligolwAPI/html/ligolw_dtd.txt':
    # if the XML file contants a http pointer to the ligolw DTD at CIT then
    # return a local copy to avoid any network problems
    return 'file://localhost' + os.path.join( os.environ["GLUE_PREFIX"],
      'etc/ligolw_dtd.txt' )
  else:
    # otherwise just use the uri in the file
    return uri


# use a local copy of the DTD, if one is available
try:
    GLUE_PREFIX = os.environ["GLUE_PREFIX"]
    xmlparser.eoCB = dtd_uri_callback
    logger.info("Using local DTD in " + 
      'file://localhost' + os.path.join( GLUE_PREFIX, 'etc') )
except KeyError:
    logger.warning('GLUE_PREFIX not set, unable to use local DTD') 

# initialize dictionaries for the dmt processes and segments definers
dmt_proc_dict = {}
dmt_seg_def_dict = {}
creator_db = None

class Server(object):
  """
  """
  def __init__(self):
    """
    """

    # create a selector to use for dispatching
    mySelector = selector.Selector()
    self.mySelector = mySelector

    # for convenience
    global configuration
    self.configuration = configuration

    # define dispatches
    mySelector.add('/ldbd/ping[.{format}]', POST=self.ping, GET=self.ping)
    mySelector.add('/ldbd/query[.{format}]', POST=self.query)
    mySelector.add('/ldbd/insert[.{format}]', POST=self.insert)
    mySelector.add('/ldbd/insertmap[.{format}]', POST=self.insertmap)
    mySelector.add('/ldbd/insertdmt[.{format}]', POST=self.insertdmt)

  def __call__(self, environ, start_response):
    """
    """

    # necessary for version 0.8.11 of selector because it expects to
    # use PATH_INFO but mod_wsgi is using the more modern and standard
    # SCRIPT_NAME
    environ['PATH_INFO'] = environ['SCRIPT_NAME'] + environ['PATH_INFO']

    # use selector to choose and execute the correct function 
    return self.mySelector(environ, start_response)

  def checkAuthorizationGridMap(self, environ, mapfile = None):
        """
        """
        # first make sure the client was verified
        verification = environ['SSL_CLIENT_VERIFY']
        if verification != 'SUCCESS':
            return (False, None)

        # grab the client cert
        clientCert = M2Crypto.X509.load_cert_string(environ['SSL_CLIENT_CERT'])
        clientSubject = clientCert.get_subject().__str__()
        logger.debug("The client cert subject is %s" % clientSubject)

        # see if the client cert is a proxy
        try:
            clientCert.get_ext("proxyCertInfo")
            clientCertProxy = True
            logger.debug("The client cert is a proxy")
        except LookupError:
            clientCertProxy = False
            logger.debug("The client cert is not a proxy")

        if not clientCertProxy:
            subject = clientSubject

        else:
            # find the number of certificates in the client chain
            max = None
            for e in environ.keys():
                m = re.search('SSL_CLIENT_CERT_CHAIN_(\d+)', e)
                if m:
                    logger.debug("Found SSL_CLIENT_CERT_CHAIN_%s" % m.group(1))
                    n = int(m.group(1))
                    if n > max: max = n

            certChainLength = max + 1

            logger.debug("There are %d certs in the chain" % certChainLength)
                
            # grab each certificate in the chain
            chainCerts = [ M2Crypto.X509.load_cert_string(environ['SSL_CLIENT_CERT_CHAIN_%d' % i]) for i in range(certChainLength) ]

            # walk the chain and find the first end entity certificate
            logger.debug("Walking the cert chain now...")
            for c in chainCerts:
                s = c.get_subject().__str__()
                logger.debug("Chain cert subject is %s" % s)
                try:
                    c.get_ext("proxyCertInfo")
                    logger.debug("Chain cert %s is a proxy" % s)
                except LookupError:
                    logger.debug("Chain cert %s is not a proxy" % s)
                    break
            subject = s

        logger.debug("Authorizing against %s" % subject)

        # parse the grid-mapfile and see if subject in it
        authorized = False
        if not mapfile:
            mapfile = configuration['gridmap']
        try:
            g = GridMap(mapfile)
            if subject in g:
                authorized = True
            else:
                authorized = False
        except Exception as e:
            logger.error("Unable to check authorization in grid-mapfile %s: %s" % (mapfile, e))

        return (authorized, subject)

  def ping(self, environ, start_response):
    """
    """
    logger.debug("Method ping called")

    # determine protocol
    try:
        protocol = decode(environ['wsgi.input'].read())
    except Exception as e:
        if (str(e)).strip() == "empty JSON description":
           logger.debug("No protocol given, request may come from Web GUI")
           protocol = "https"
        else:
           start_response("400 Bad Request", [('Content-type', 'text/plain')])
           msg = "400 Bad Request"
           logger.debug("Error decoding input: %s" % e)
           return [ msg ]


    if protocol == "https": 
      # use generic grid-mapfile for ping operation
      mapfile = self.configuration['gridmap']

      # check authorization
      authorized, subject = self.checkAuthorizationGridMap(environ, mapfile)
      if not authorized:
        start_response("401 Unauthorized", [('Content-type', 'text/plain')])
        msg = "401 Unauthorized\n\nSubject %s is not authorized for method ping" % subject
        logger.info("Subject %s is not authorized for method ping" % subject)
        return [ msg ]
      else:
        logger.info("Subject %s is authorized for method ping" % subject)
    else:
      logger.info("Method ping is being called from internal network")


    try:
      hostname = socket.getfqdn()
      msg = "LDBD server at %s is alive" % hostname
    except Exception as e:
      msg = "LDBD server is alive" 

    format = environ['wsgiorg.routing_args'][1]['format']
    if format == 'json':
        result = encode(msg)
        contentType = 'application/json'
    else:
        result = msg
        contentType = 'text/plain'

    header = [('Content-Type', contentType)]
    start_response("200 OK", header)

    return [ result ]

  def query(self, environ, start_response):
    """
    """
    global xmlparser, dbobj

    logger.debug("Method query called")

    # determine protocol
    try:
        protocol, querystr = (decode(environ['wsgi.input'].read())).split(":")
    except Exception as e:
        start_response("400 Bad Request", [('Content-type', 'text/plain')])
        msg = "400 Bad Request"
        logger.debug("Error decoding input: %s" % e)
        return [ msg ]


    if protocol == "https":
      # use generic grid-mapfile for query operation
      mapfile = self.configuration['gridmap']

      # check authorization
      authorized, subject = self.checkAuthorizationGridMap(environ, mapfile)
      if not authorized:
        start_response("401 Unauthorized", [('Content-type', 'text/plain')])
        msg = "401 Unauthorized\n\nSubject %s is not authorized for method query" % subject
        logger.info("Subject %s is not authorized for method query" % subject)
        return [ msg ]
      else:
        logger.info("Subject %s is authorized for method query" % subject)
    else:
      logger.info("Method QUERY is being called from internal network")


    # pick off the format input type
    format = environ['wsgiorg.routing_args'][1]['format']

    # for a POST the format must be JSON
    if format != 'json':
        start_response("400 Bad Request", [('Content-type', 'text/plain')])
        msg = "400 Bad Request\n\nformat must be 'json' for POST operation"
        return [ msg ]

    logger.debug("Method query called with '%s'" % querystr)

    try:
      # create a ligo metadata object
      lwtparser = ldbd.LIGOLwParser()
      ligomd = ldbd.LIGOMetadata(xmlparser,lwtparser,dbobj)

      # execute the query
      rowcount = ligomd.select(querystr)

      # convert the result to xml
      result = ligomd.xml()

      logger.debug("Method query: %d rows returned" % rowcount)
    except Exception as e:
      start_response("500 Internal Server Error", [('Content-type', 'text/plain')])
      msg = "500 Internal Server Error\n\n%s" % e
      logger.error(msg)
      return [ msg ]

    try:
      del ligomd
      del lwtparser
    except Exception as e:
      logger.error("Error deleting metadata object in method query: %s" % e)

    # encode the result
    result = encode(result)
    
    # return the result
    header = [('Content-Type', 'application/json')]
    start_response("200 OK", header)

    return [ result ]

  def insert(self, environ, start_response):
    """
    """
    global xmlparser, dbobj

    logger.debug("Method insert called")

    # use specific grid-mapfile for insert operation
    mapfile = self.configuration['gridmap_insert']

    # check authorization
    authorized, subject = self.checkAuthorizationGridMap(environ, mapfile)
    if not authorized:
        start_response("401 Unauthorized", [('Content-type', 'text/plain')])
        msg = "401 Unauthorized\n\nSubject %s is not authorized for method insert" % subject
        logger.info("Subject %s is not authorized for method insert" % subject)
        return [ msg ]
    else:
        logger.info("Subject %s is authorized for method insert" % subject)

    # pick off the format input type
    format = environ['wsgiorg.routing_args'][1]['format']

    # for a POST the format must be JSON
    if format != 'json':
        start_response("400 Bad Request", [('Content-type', 'text/plain')])
        msg = "400 Bad Request\n\nformat must be 'json' for POST operation"
        return [ msg ]

    # read the incoming payload
    try:
        #import simplejson  (moved to top)
        wsgiIn=environ['wsgi.input'].read()
        inputString=simplejson.loads(wsgiIn)
        #inputString = decode(environ['wsgi.input'].read())
    except Exception as e:
        start_response("400 Bad Request", [('Content-type', 'text/plain')])
        msg = "400 Bad Request"
        logger.debug("Error decoding input: %s" % e)
        return [ msg ]

    logger.debug("Method insert called with '%s'" % inputString)

    try:
      # create a ligo metadata object
      lwtparser = ldbd.LIGOLwParser()
      ligomd = ldbd.LIGOMetadata(xmlparser,lwtparser,dbobj)

      # parse the input string into a metadata object
      ligomd.parse(inputString)

      # add a gridcert table to this request containing the users dn
      ligomd.set_dn(subject)

      # insert the metadata into the database
      result = str(ligomd.insert())

      logger.info("Method insert: %s rows affected by insert" % result)
    except Exception as e:
      start_response("500 Internal Server Error", [('Content-type', 'text/plain')])
      msg = "500 Internal Server Error\n\n%s" % e
      logger.error(msg)
      return [ msg ]

    try:
      del ligomd
      del lwtparser
    except Exception as e:
      logger.error("Error deleting metadata object in method query: %s" % e)

    # encode the result
    result = encode(result)
    
    # return the result
    header = [('Content-Type', 'application/json')]
    start_response("200 OK", header)

    return [ result ]
    
  def insertmap(self, environ, start_response):
    """
    """
    logger.debug("Method insertmap called")

    # use specific grid-mapfile for insert operation
    mapfile = self.configuration['gridmap_insertmap']

    # check authorization
    authorized, subject = self.checkAuthorizationGridMap(environ, mapfile)
    if not authorized:
        start_response("401 Unauthorized", [('Content-type', 'text/plain')])
        msg = "401 Unauthorized\n\nSubject %s is not authorized for method insertmap" % subject
        logger.info("Subject %s is not authorized for method insertmap" % subject)
        return [ msg ]
    else:
        logger.info("Subject %s is authorized for method insertmap" % subject)

    # pick off the format input type
    format = environ['wsgiorg.routing_args'][1]['format']

    # for a POST the format must be JSON
    if format != 'json':
        start_response("400 Bad Request", [('Content-type', 'text/plain')])
        msg = "400 Bad Request\n\nformat must be 'json' for POST operation"
        return [ msg ]

    start_response("500 Internal Server Error", [('Content-type', 'text/plain')])
    msg = "500 Internal Server Error\n\nserver is not initialized for RLS connections"
    logger.error(msg)
    return [ msg ]

  def insertdmt(self, environ, start_response):
    """
    """
    global xmlparser, dbobj
    global dmt_proc_dict, dmt_seg_def_dict, creator_db
    proc_key = {}
    known_proc = {}
    seg_def_key = {}

    logger.debug( "Method dmtinsert called." )
    logger.debug( "Known processes %s, " % str(dmt_proc_dict) )
    logger.debug( "Known segment_definers %s" % str(dmt_seg_def_dict) )

    # use specific grid-mapfile for insertdmt operation
    mapfile = self.configuration['gridmap_insertdmt']

    # check authorization
    authorized, subject = self.checkAuthorizationGridMap(environ, mapfile)
    if not authorized:
        start_response("401 Unauthorized", [('Content-type', 'text/plain')])
        msg = "401 Unauthorized\n\nSubject %s is not authorized for method insertdmt" % subject
        logger.info("Subject %s is not authorized for method insertdmt" % subject)
        return [ msg ]
    else:
        logger.info("Subject %s is authorized for method insertdmt" % subject)

    # pick off the format input type
    format = environ['wsgiorg.routing_args'][1]['format']

    # for a POST the format must be JSON
    if format != 'json':
        start_response("400 Bad Request", [('Content-type', 'text/plain')])
        msg = "400 Bad Request\n\nformat must be 'json' for POST operation"
        return [ msg ]

    # read the incoming payload
    try:
        inputString = decode(environ['wsgi.input'].read())
    except Exception as e:
        start_response("400 Bad Request", [('Content-type', 'text/plain')])
        msg = "400 Bad Request"
        logger.debug("Error decoding input: %s" % e)
        return [ msg ]

    logger.debug("Method insertdmt called with '%s'" % inputString)

    try:
      # create a ligo metadata object
      lwtparser = ldbd.LIGOLwParser()
      ligomd = ldbd.LIGOMetadata(xmlparser,lwtparser,dbobj)

      # parse the input string into a metadata object
      logger.debug("parsing xml data")
      ligomd.parse(inputString)

      # store the users dn in the process table
      ligomd.set_dn(subject)

      # determine the local creator_db number
      if creator_db is None:
        sql = "SELECT DEFAULT FROM SYSCAT.COLUMNS WHERE "
        sql += "TABNAME = 'PROCESS' AND COLNAME = 'CREATOR_DB'"
        ligomd.curs.execute(sql)
        creator_db = ligomd.curs.fetchone()[0]

      # determine the locations of columns we need in the process table
      process_cols = ligomd.table['process']['orderedcol']
      node_col = process_cols.index('node')
      prog_col = process_cols.index('program')
      upid_col = process_cols.index('unix_procid')
      start_col = process_cols.index('start_time')
      end_col = process_cols.index('end_time')
      pid_col = process_cols.index('process_id')

      # determine and remove known entries from the process table
      rmv_idx = []
      for row_idx,row in enumerate(ligomd.table['process']['stream']):
        uniq_proc = (row[node_col],row[prog_col],row[upid_col],row[start_col])
        try:
          proc_key[str(row[pid_col])] = dmt_proc_dict[uniq_proc]
          known_proc[str(dmt_proc_dict[uniq_proc])] = row[end_col]
          rmv_idx.append(row_idx)
        except KeyError:
          # we know nothing about this process, so query the database
          sql = "SELECT BLOB(process_id) FROM process WHERE "
          sql += "creator_db = " + str(creator_db) + " AND "
          sql += "node = '" + row[node_col] + "' AND "
          sql += "program = '" + row[prog_col] + "' AND "
          sql += "unix_procid = " + str(row[upid_col]) + " AND "
          sql += "start_time = " + str(row[start_col])
          ligomd.curs.execute(sql)
          db_proc_ids = ligomd.curs.fetchall()
          if len(db_proc_ids) == 0:
            # this is a new process with no existing entry
            dmt_proc_dict[uniq_proc] = row[pid_col]
          elif len(db_proc_ids) == 1:
            # the process_id exists in the database so use that insted
            dmt_proc_dict[uniq_proc] = db_proc_ids[0][0]
            proc_key[str(row[pid_col])] = dmt_proc_dict[uniq_proc]
            known_proc[str(dmt_proc_dict[uniq_proc])] = row[end_col]
            rmv_idx.append(row_idx)
          else:
            # multiple entries for this process, needs human assistance
            raise ServerHandlerException("multiple entries for dmt process")

      # delete the duplicate processs rows and clear the table if necessary
      newstream = []
      for row_idx,row in enumerate(ligomd.table['process']['stream']):
        try:
          rmv_idx.index(row_idx)
        except ValueError:
          newstream.append(row)
      ligomd.table['process']['stream'] = newstream
      if len(ligomd.table['process']['stream']) == 0:
        del ligomd.table['process']

      # delete the duplicate process_params rows and clear the table if necessary
      # (the DMT does not write a process_params table, so check for one first)
      if 'process_params' in ligomd.table:
        ppid_col = ligomd.table['process_params']['orderedcol'].index('process_id')
        newstream = []
        for row_idx,row in enumerate(ligomd.table['process_params']['stream']):
          # if the process_id in this row is known, delete (i.e. don't copy) it
          try:
            proc_key[str(row[ppid_col])]
          except KeyError:
            newstream.append(row)
        ligomd.table['process_params']['stream'] = newstream
        if len(ligomd.table['process_params']['stream']) == 0:
          del ligomd.table['process_params']

      # turn the known process_id binary for this insert into ascii
      for pid in known_proc.keys():
        pid_str = "x'"
        for ch in pid:
          pid_str += "%02x" % ord(ch)
        pid_str += "'"
        known_proc[pid] = (pid_str, known_proc[pid])

      # determine the locations of columns we need in the segment_definer table
      seg_def_cols = ligomd.table['segment_definer']['orderedcol']
      ifos_col = seg_def_cols.index('ifos')
      name_col = seg_def_cols.index('name')
      vers_col = seg_def_cols.index('version')
      sdid_col = seg_def_cols.index('segment_def_id')

      # determine and remove known entries in the segment_definer table
      rmv_idx = []
      for row_idx,row in enumerate(ligomd.table['segment_definer']['stream']):
        uniq_def = (row[ifos_col],row[name_col],row[vers_col])
        try:
          seg_def_key[str(row[sdid_col])] = dmt_seg_def_dict[uniq_def]
          rmv_idx.append(row_idx)
        except KeyError:
          # we know nothing about this segment_definer, so query the database
          sql = "SELECT BLOB(segment_def_id) FROM segment_definer WHERE "
          sql += "creator_db = " + str(creator_db) + " AND "
          sql += "ifos = '" + row[ifos_col] + "' AND "
          sql += "name = '" + row[name_col] + "' AND "
          sql += "version = " + str(row[vers_col])
          ligomd.curs.execute(sql)
          db_seg_def_id = ligomd.curs.fetchall()
          if len(db_seg_def_id) == 0:
            # this is a new segment_defintion with no existing entry
            dmt_seg_def_dict[uniq_def] = row[sdid_col]
          else:
            dmt_seg_def_dict[uniq_def] = db_seg_def_id[0][0]
            seg_def_key[str(row[sdid_col])] = dmt_seg_def_dict[uniq_def]
            rmv_idx.append(row_idx)

      # delete the necessary rows. if the table is empty, delete it
      newstream = []
      for row_idx,row in enumerate(ligomd.table['segment_definer']['stream']):
        try:
          rmv_idx.index(row_idx)
        except ValueError:
          newstream.append(row)
      ligomd.table['segment_definer']['stream'] = newstream
      if len(ligomd.table['segment_definer']['stream']) == 0:
        del ligomd.table['segment_definer']

      # now update the values in the xml with the values we know about
      for tabname in ligomd.table.keys():
        table = ligomd.table[tabname]
        if tabname == 'process':
          # we do nothing to the process table
          pass
        elif tabname == 'segment' or tabname == 'segment_summary':
          # we need to update the process_id and the segment_def_id columns
          pid_col = table['orderedcol'].index('process_id')
          sdid_col = table['orderedcol'].index('segment_def_id')
          row_idx = 0
          for row in table['stream']:
            try:
              repl_pid = proc_key[str(row[pid_col])]
            except KeyError:
              repl_pid = row[pid_col]
            try:
              repl_sdid = seg_def_key[str(row[sdid_col])]
            except KeyError:
              repl_sdid = row[sdid_col]
            row = list(row)
            row[pid_col] = repl_pid
            row[sdid_col] = repl_sdid
            table['stream'][row_idx] = tuple(row)
            row_idx += 1
        else:
          # we just need to update the process_id column
          pid_col = table['orderedcol'].index('process_id')
          row_idx = 0
          for row in table['stream']:
            try:
              repl_pid = proc_key[str(row[pid_col])]
              row = list(row)
              row[pid_col] = repl_pid
              table['stream'][row_idx] = tuple(row)
            except KeyError:
              pass
            row_idx += 1

      # insert the metadata into the database
      logger.debug("inserting xml data")
      result = str(ligomd.insert())
      logger.debug("insertion complete")

      # update the end time of known processes in the process table
      for pid in known_proc.keys():
        # first check to see if we are backfilling missing segments
        sql = "SELECT end_time,domain FROM process "
        sql += " WHERE process_id = " + known_proc[pid][0]
        ligomd.curs.execute(sql)
        last_end_time = ligomd.curs.fetchone()

        # check the dn in the row we are about to update matches the users dn
        dn = last_end_time[1].strip()
        if subject != dn:
          msg = "\"%s\" does not match dn in existing row entries: " % subject
          msg += "%s (process_id %s)" % (dn, known_proc[pid][0])
          logger.warn(msg)
        else:
          logger.debug('"%s" updating process_id %s' % (dn, known_proc[pid][0]))

        if int(known_proc[pid][1]) <= int(last_end_time[0]):
          logger.debug("Backfilling missing segments for process_id " +
            known_proc[pid][0] + " not updating end_time")
        else:
          # if we are not backfilling, update the end_time of the process
          sql = "UPDATE process SET end_time = " + str(known_proc[pid][1])
          sql += " WHERE process_id = " + known_proc[pid][0]
          sql += " AND end_time < " + str(known_proc[pid][1])
          ligomd.curs.execute(sql)
      ligomd.dbcon.commit()

      logger.info("Method insert: %s rows affected by insert" % result)

    except Exception as e:
      start_response("500 Internal Server Error", [('Content-type', 'text/plain')])
      msg = "500 Internal Server Error\n\n%s" % e
      logger.error(msg)
      return [ msg ]

    try:
      del ligomd
      del lwtparser
      del known_proc
      del seg_def_key
      del proc_key
    except Exception as e:
      logger.error("Error deleting metadata object in method insertdmt: %s" % e)

    # encode the result
    result = encode(result)
    
    # return the result
    header = [('Content-Type', 'application/json')]
    start_response("200 OK", header)

    return [ result ]

class GridMapError(Exception):
    """
    Raised for errors in GridMap class
    """
    pass

class GridMap(dict):
    """
    """
    def __init__(self, path=None):
        """
        """
        self.path = path

        # initialize the base class
        dict.__init__(self)

        if not path:
            msg = "No path for grid-mapfile"
            raise GridMapError(msg)

    def parse(self):
        """
        """
        # clear any existing entries
        self.clear()

        # parse the grid-mapfile
        try:
            f = open(self.path, 'r')
        except Exception as e:
            msg = "Unable to open %s for reading: %s" % (self.path, e)
            raise GridMapError(msg)

        for line in f:
            s = line.strip().split('"')
            if len(s) == 3: subject = s[1]
            elif len(s) == 2: subject = s[1]
            elif len(s) == 1: subject = s[0]
            else:
                msg = "Error parsing line %s" % line
                raise GridMapError(msg)

            dict.__setitem__(self, subject, 1)

        f.close()

    def __getitem__(self, key):
        """
        """
        self.parse()

        return dict.__getitem__(self, key)

    def has_key(self, key):
        """
        """
        self.parse()

        return dict.has_key(self, key)

    def keys(self):
        """
        """
        self.parse()

        return dict.keys(self)





# script to extract data quality segments from GEO h(t) frame files
# and output ligolw XML suitable for pubishing to the segdb
#
# Copyright (C) 2010 Peter Couvares
# 
# This is part of the Grid LSC User Environment (GLUE)
# 
# GLUE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
# 
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

# TODO: convert debugging output from print() to logger()

from __future__ import print_function
import os
import sys
import time
import pwd

from optparse import OptionParser

from ligo import segments
from glue import lal
from ligo.segments import utils as segmentsUtils
from glue import gpstime

import glue.pipeline
import glue.utils

from glue.ligolw import ligolw
from glue.ligolw import table
from glue.ligolw import lsctables
from glue.ligolw import utils as ligolw_utils
from glue.ligolw.utils import process as ligolw_process
from glue.segmentdb import segmentdb_utils
from glue import pidfile as pidfile

import pylal.Fr

from pylal import git_version


### constants ###

# mapping of 12-bit GEO data-quality word to DQ flags
# based on http://www.geo600.uni-hannover.de/geodc/index.php?location=data
dq_word_key = [
    'GEO-UNLOCKED',
    'GEO-HW_MAINT_ON',
    'GEO-SW_MAINT_ON',
    'GEO-CALIB_BROKEN',
    'GEO-CONFIG_REREAD',
    'GEO-TIMESTAMP_BROKEN',
    'GEO-CHI2_GT_5',
    'GEO-CHI2_GT_6',
    'GEO-CHI2_GT_7',
    'GEO-CHI2_GT_8',
    'GEO-CHI2_GT_9',
    'GEO-CHI2_GT_10',
    ]

__author__  = "Peter Couvares <pfcouvar@syr.edu>"

PROGRAM_NAME = os.path.abspath(sys.argv[0])


class diskcacheEntry(object):
    """
    An object representing one line in a diskcache file.
    """

    def __init__(self, *args, **kwargs):
        """
        Intialize a diskcacheEntry object.  Takes a single string
        argument, which is interpreted and parsed as a line from a
        diskcache file.
        """

        if len(args) != 1:
            raise TypeError("invalid arguments: %s" % args)

        line = args[0]
        self.__frame_segments__ = None
        self.__frame_files__ = None
        # parse line of text as an entry in a cache file
        try:
            # split on spaces and then comma to get the parts
            header, self.modTime, self.fileCount, times = line.strip().split(' ', 3)
            self.dir, self.site, self.frameType, self.frameCount, self.duration = header.split(',')
            self.fileCount = int(self.fileCount)
            self.modTime = int(self.modTime)
            self.frameCount = int(self.frameCount)
            self.duration = int(self.duration)

            # times string has form { t1 t2 t3 t4 t5 t6 ... tN t(N+1) }
            # where the (ti, t(i+1)) represent segments
            #
            # first turn the times string into a list of integers
            times = [ int(s) for s in times[1:-1].strip().split(' ') ]

            # group the integers by two and turn those tuples into segments
            assert len(times) % 2 is 0
            self.segments = segments.segmentlist([ segments.segment(t) for t in zip(times[::2],times[1::2]) ])

        except Exception as e:
            msg = "Error parsing frame cache file with line %s: %s" % (line, e)
            raise ValueError(msg)
            return


    @property
    def frame_segments(self):
        """
        Return an "uncoalesced" segmentlist containing one segment for
        each frame file.
        """
        if self.__frame_segments__:
            return self.__frame_segments__
        self.__frame_segments__ = segments.segmentlist([])
        for seg in self.segments:
            self.__frame_segments__.extend(segmentsUtils.segmentlist_range(seg[0], seg[1], self.duration))
        return self.__frame_segments__


    @property
    def frame_files(self):
        """
        Return a list containing the physical file name of each frame.
        """
        if self.__frame_files__:
            return self.__frame_files__
        self.__frame_files__ = []
        for seg in self.frame_segments:
            frame_filename = "%s-%s-%s-%s.gwf" % (self.site, self.frameType, seg[0], self.duration)
            pfn = os.path.join(self.dir, frame_filename)
            self.__frame_files__.append(pfn)
        return self.__frame_files__


"""
Given a state file and GPS time range, returns all segments within
that range for which GEO frame files have *not* been processed into DQ
XML files.

NOTE: the state file contains no calibration version or segment
version information -- it's assumed you will specify a new state file
if/when you want to process a newer version of any frame files, since
the segments may overlap with previous versions.
"""
def get_pending_segments(state_file, gps_start, gps_end, verbose=False):
    if verbose: print("reading state from %s" % state_file)
    # load our state file and extract processed/excluded segment lists
    indoc = ligolw_utils.load_url(state_file)
    processed_segments = segmentdb_utils.find_segments(indoc,"P1:PROCESSED:0")
    exclude_segments = segmentdb_utils.find_segments(indoc,"P1:EXCLUDE:0")
    if verbose: print("processed segments =", processed_segments)
    if verbose: print("excluded segments =", exclude_segments)
    # calculate segments to be processed
    gps_range = segments.segmentlist([segments.segment(gps_start, gps_end)])
    if verbose: print("gps range =", gps_range)
    pending_segments = (gps_range - processed_segments) - exclude_segments
    if verbose:
        print("pending segments within %s = %s" % (gps_range, pending_segments))
    return pending_segments


# TODO: refactor get_pending_segments() and append_processed_segments()
# to use some kind of common state_file object, so we don't have open & read
# the same state file in each

"""
Given a state file and segmentlist, add the segments to the table of
processed segments.
"""
def append_processed_segments(state_file, segmentlist, exclude_segmentlist, verbose=False):
    if verbose: print("appending segmentlist %s to state file %s" % (segmentlist, state_file))
    if verbose: print("appending exclude_segmentlist %s to state file %s" % (segmentlist, state_file))
    # load existing state file and extract processed/excluded segment lists
    indoc = ligolw_utils.load_url(state_file)
    processed_segments = segmentdb_utils.find_segments(indoc,"P1:PROCESSED:0")
    exclude_segments = segmentdb_utils.find_segments(indoc,"P1:EXCLUDE:0")
    # construct new state file with updated segments
    outdoc = ligolw.Document()
    outdoc.appendChild(ligolw.LIGO_LW())
    proc_id = ligolw_process.register_to_xmldoc(outdoc, PROGRAM_NAME, options.__dict__).process_id
    excl_def_id = segmentdb_utils.add_to_segment_definer(outdoc,proc_id,"P1","EXCLUDE",0)
    pub_def_id = segmentdb_utils.add_to_segment_definer(outdoc,proc_id,"P1","PROCESSED",0)
    if verbose: print("previously processed segments = %s" % processed_segments)
    # append new segments
    processed_segments += segmentlist
    exclude_segments += exclude_segmentlist
    # Do I need to explicity coalesce here?  Let's do so, just to be safe.
    processed_segments.coalesce()
    exclude_segments.coalesce()
    if verbose: print("new processed segments = %s" % processed_segments)
    if verbose: print("new excluded segments = %s" % exclude_segments)
    segmentdb_utils.add_to_segment(outdoc,proc_id,excl_def_id,exclude_segments)
    segmentdb_utils.add_to_segment(outdoc,proc_id,pub_def_id,processed_segments)
    ligolw_utils.write_filename(outdoc, state_file)


"""
Given a GEO h(t) frame file and optional DQ word bitmask, return two
name->segmentlist hashes: one containing DQ summary segments and one
containing the DQ segments themselves.

The default bitmask (4032 or int('111111000000',2)) omits the first
six low-order bits, since they are summarized by the GEO-SCIENCE flag.
(The GEO-SCIENCE flag is always included in the results, regardless of
the bitmask.)
"""
def get_dq_flags(file, ifo, bitmask=4032, verbose=False):
    # extract the DQ state vector from the frame file
    state_vec = pylal.Fr.frgetvect(file, ifo + ':DER_DATA_QUALITY', start=-1,
                                   span=-1, verbose=verbose)
    # read the frame's start time and duration
    gps_start = int(state_vec[1])
    num_frames = len(state_vec[0])
    # TODO: frame_len should be pulled from the frame data, not hardcoded
    frame_len = 1
    total_len = num_frames * frame_len

    if verbose: print("gps_start =", gps_start, "num_frames =", num_frames)

    # initialize name->segmentlist hashes for summary segments and segments
    summary_segs = {}
    segs = {}

    # summary segments need to be created for any time in which a
    # segment's state is known, regardless of its value -- so by
    # virtue of a frame's existence, there should be summary segments
    # for the entire time it covers, for all segments we want to
    # record (always GEO-SCIENCE, plus any others specified in the
    # bitmask)

    total_seg = segments.segment(gps_start, gps_start + total_len)
    summary_segs['GEO-SCIENCE'] = segments.segmentlist([total_seg])
    segs['GEO-SCIENCE'] = segments.segmentlist()
    for i, name in enumerate(dq_word_key):
        if (bitmask >> i) & 1:
            summary_segs[name] = segments.segmentlist([total_seg])
            # also initialize segs hash for later
            segs[name] = segments.segmentlist()

    # now we need to populate the actual segments seen in the frames...
    for frame_i in range(0,num_frames):
        frame_start = gps_start + (frame_i * frame_len)
        frame_end = frame_start + frame_len
        frame_seg = segments.segment(frame_start, frame_end)

        # if all of the first six bits are zero, GEO-SCIENCE is True
        if (state_vec[0][frame_i] & int('111111',2)) == 0:
            segs['GEO-SCIENCE'].append(frame_seg)

        # generate any segments specified in bitmask
        for i, name in enumerate(dq_word_key):
            # if bitmask includes this bit AND it's set in the frame,
            # create corresponding segment
            if ((bitmask >> i) & 1) and ((state_vec[0][frame_i] >> i) & 1):
                segs[name].append(frame_seg)

    return (gps_start, total_len, summary_segs, segs)


def gen_dq_xml_filename(xml_root, ifo, gps_start, duration):
    """
    Given an XML root dir, IFO name, GPS start time, and frame duration,
    returns the full-path data-quality XML filename into which the
    segments should be written, following our (undocumented?) DQ XML
    directory and file-naming conventions.
    """
    subdir = "%c-DQ_Segments-%d" % (ifo[0], gps_start/100000)
    file = "%c-DQ_Segments-%d-%d.xml" % (ifo[0], gps_start, duration)
    return os.path.join(xml_root, subdir, file)


def frame_to_dq_xml(frame_file, xml_root, ifo, type, version, comment, verbose=False):
    """
    Given a frame file, output xml root dir, and various frame
    attributes, write a ligolw XML file containing the frame's DQ
    segments.
    """
    # extract the frame's DQ info
    (gps_start, duration, summary_segs, segs) = get_dq_flags(frame_file, ifo, verbose=verbose)

    # create a blank xml document and add the process id
    outdoc = ligolw.Document()
    outdoc.appendChild(ligolw.LIGO_LW())
    proc_id = ligolw_process.register_to_xmldoc(outdoc, PROGRAM_NAME, options.__dict__, comment=comment, ifos=[ifo], version=git_version.id, cvs_repository=git_version.branch, cvs_entry_time=git_version.date).process_id

    # create a segment_definer for each flag
    for flag in summary_segs:
        summary_segs[flag].coalesce()
        if verbose: print("flag =", flag, "\nsums =", summary_segs[flag])
        seg_def_id = segmentdb_utils.add_to_segment_definer(outdoc,proc_id,ifo,flag,version,comment=comment)
        segmentdb_utils.add_to_segment_summary(outdoc,proc_id,seg_def_id,summary_segs[flag],comment=type)
        if flag in segs:
            segs[flag].coalesce()
            if verbose: print("segs =", segs[flag])
            segmentdb_utils.add_to_segment(outdoc,proc_id,seg_def_id,segs[flag])

    # write the xml doc to disk
    proctable = lsctables.ProcessTable.get_table(outdoc)
    proctable[0].end_time = gpstime.GpsSecondsFromPyUTC(time.time())
    outname = gen_dq_xml_filename(xml_root, ifo, gps_start, duration)
    glue.utils.mkdir_p(os.path.dirname(outname))
    ligolw_utils.write_filename(outdoc, outname)

    return segments.segment(gps_start, gps_start + duration)


### option parsing ###

parser = OptionParser(
    version = git_version.verbose_msg,
    usage   = "%prog [OPTIONS]",
    description = "Extracts data quality segments from GEO h(t) frame files, and writes it to XML." )

# defaults are set for use on segdb.ligo.caltech.edu (circa July 2010)
parser.add_option('-s', '--state-file', metavar='FILE', help='state file')
parser.add_option('-c', '--cache-file', metavar='FILE', help='diskCache file')
parser.add_option('-x', '--xml-root', metavar='DIR', help='XML output directory',
                  default='/archive/frames/online/DQ')
parser.add_option('-l', '--lock_dir', metavar='DIR', help='lock file directory (use with caution)')
parser.add_option("-t", '--frame-type', metavar='RDS_TYPE', help='RDS frame type',
                  default='G1_RDS_C01_L3' )
parser.add_option('-v', '--segment-version', metavar="VERSION",
                  help="create segments as version number VERSION", type="int",
                  default=1)
parser.add_option('-i', '--ifo', metavar='IFO', default='G1' )
parser.add_option('--gps-start', metavar='GPS_TIME', help='GPS start time',
                  type="int", default=0 )
parser.add_option('--gps-end', metavar='GPS_TIME', help='GPS end time',
                  type="int", default=gpstime.GpsSecondsFromPyUTC(time.time()) )
parser.add_option("-C", "--comment", metavar="STRING",
                  help="add the optional STRING as the process:comment",
                  default='' )
parser.add_option("-V", "--verbose", action="store_true",
                  help="print extra debugging information",
                  default=False )

options, argv_frame_files = parser.parse_args()

if options.cache_file:
    if len(argv_frame_files) > 0:
        raise ValueError("individual frame files cannot be specified along with a diskCache")
    if not os.path.isfile(options.cache_file):
        raise ValueError("diskCache file not found (%s)" % options.cache_file)
else:
    if len(argv_frame_files) == 0:
        raise ValueError("no diskCache or frame files specified")

if options.state_file and not os.path.isfile(options.state_file):
    raise ValueError("state file not found (%s)" % options.state_file)
if not os.path.isdir(options.xml_root):
    raise ValueError("XML root directory not found (%s)" % options.xml_root)

xml_root = options.xml_root.rstrip(os.sep) + os.sep + options.ifo
if not os.path.isdir(xml_root):
    # be forgiving if the user specified "xml_root/ifo" instead of just xml_root
    if options.xml_root.rstrip(os.sep).endswith(os.sep + options.ifo) and os.path.isdir(options.xml_root):
        xml_root = options.xml_root
    else:
        raise ValueError("XML root dir '%s' does not contain necessary IFO subdir '%s'" % (options.xml_root, options.ifo))

### main ###

"""
NOTE: to be really safe, if any of the data (the frames, DQ files,
statefile, etc.) are on a shared filesystem, then the lockfile should
be kept on that same filesystem (rather than TMPDIR) so that locking
has a chance of working if two instances of the script are run on
different hosts.  In our current use this isn't a concern.
"""
# grab the lockfile
if options.lock_dir is None:
    lock_dir = os.getenv('TMPDIR') or '/tmp'
lockfile = os.path.join(lock_dir, "%s.pid" % os.path.basename(PROGRAM_NAME))
pidfile.get_lock(lockfile)

# find any unprocessed frame files

pending_segments = segments.segmentlist()
if options.state_file:
    pending_segments = get_pending_segments(options.state_file,
                                            options.gps_start,
                                            options.gps_end,
                                            options.verbose)
else:
    pending_segments = segments.segmentlist([segments.segment(options.gps_start, options.gps_end)])
if options.verbose: print("pending_segments =", pending_segments)

pending_files = []
if options.cache_file:
    # find all frame files for our ifo, type, and pending_segments
    cache_f = open(options.cache_file, 'r')
    for line in cache_f:
        entry = diskcacheEntry(line)
        if (entry.frameType != options.frame_type or
            entry.site != options.ifo[0] or
            pending_segments.intersects(entry.segments) is False):
            continue
        # each cache entry corresponds to a dir containing one or more
        # frame files; at this point we know at least one of the files
        # matches, but we need to figure out exactly which ones based
        # on their segments
        for (segment, framefile) in zip(entry.frame_segments, entry.frame_files):
            if not pending_segments.intersects_segment(segment):
                continue
            if not os.path.isfile(framefile):
                if options.verbose: print("warning: %s not found" % framefile)
                True
            else:
                pending_files.append(framefile)
else:
    pending_files = argv_frame_files

if options.verbose: print("pending files =", pending_files)

newly_processed_segments = segments.segmentlist([])
corrupt_frame_segments = segments.segmentlist([])
for file in pending_files:
    try:
        segment = frame_to_dq_xml(file, options.xml_root, options.ifo,
                                  options.frame_type, options.segment_version,
                                  options.comment, verbose=options.verbose)
        newly_processed_segments += [segment]
#        assert [segment] == segmentsUtils.fromfilenames([file])
    except Exception as e:
        if (str(e).startswith(" *** FrError: in FrFileIOpen unknown format") or
            "vector not found" in str(e)):
            # add the current corrupt frame's segment to the exclude list
            corrupt_frame_segments += segmentsUtils.fromfilenames([file])
            if options.state_file:
                # update state file with newly-processed frame segments
                append_processed_segments(options.state_file, newly_processed_segments, corrupt_frame_segments, verbose=options.verbose)
        else:
            # update the state file in case we're about to die
            if options.state_file:
                append_processed_segments(options.state_file, newly_processed_segments, corrupt_frame_segments, verbose=options.verbose)
            raise Exception(e)


if options.state_file:
    # update state file with newly-processed frame segments
    append_processed_segments(options.state_file, newly_processed_segments, corrupt_frame_segments,
                              verbose=options.verbose)

os.remove(lockfile)




###################################################################
from __future__ import print_function
import os
import socket
import pwd
import re
import time
import sys
import commands
import StringIO
import binascii
from optparse import OptionParser
import six

try:
  import pyRXP
except ImportError as e:
  print("""
Error: unable to import the pyRXP module.

You must have pyRXP installed and in your PYTHONPATH to run %s.

%s
""" %(sys.argv[0], e), file=sys.stderr)

try:
  from glue import gpstime
  from glue import ldbd
  from ligo import segments
  from glue import git_version
  from glue.segmentdb import segmentdb_utils
  from glue.ligolw import table
  from glue.ligolw import lsctables
  from glue.ligolw import ligolw 
  from glue.ligolw.utils import process
  from glue.ligolw import types as ligolwtypes 
except ImportError as e:
  print("""
Error: unable to import modules from glue.

Check that glue is correctly installed and in your PYTHONPATH.

%s
""" % e, file=sys.stderr)
  sys.exit(1)
##################################################################
__author__ = "Ping Wei <piwei@physics.syr.edu>"
__date__ = git_version.date
__version__ = git_version.id
__src__ = git_version.status

PROGRAM_NAME = sys.argv[0].replace('./','')
PROGRAM_PID  = os.getpid()

try:
  USER_NAME = os.getlogin()
except:
  USER_NAME = pwd.getpwuid(os.getuid())[0]


##################################################################
#            Command Line
##################################################################
def parse_command_line():
  parser = OptionParser()
  
  parser.add_option("-p", "--ping", action = "store_true", default = None, help = "Ping the target server")
  parser.add_option("-t", "--segment-url", action = "store", metavar = "URL", default = None, \
                    help = "Users have to specify protocol 'https://' for a secure connection in the segment database url. For example, '--segment-url=https://segdb.ligo.caltech.edu'. No need to specify port number'.")
  parser.add_option("-o", "--output", action = "store", metavar = "FILE", default = None, \
                    help = "Write segments to FILE rather than the segment database")
  parser.add_option("-j", "--identity", action = "store", metavar = "IDENTITY", default = None, \
                    help = "Set the subject line of the server's service certificate to IDENTITY")
  parser.add_option("-I", "--insert", action = "store_true", default = 0, help = "Insert segments to the segment database")
  parser.add_option("-A", "--append", action = "store_true", default = 0, help = "Append segments to an existing segment type")
  parser.add_option("-F", "--ignore-append-check", action = "store_true", default = 0,
                    help = "Allow --append to insert segments prior to the latest existing segment.")
  parser.add_option("-i", "--ifos", action = "store", metavar = "IFOS", default = None, \
                    help = "Set the segment interferometer to IFOS (e.g. H1)")
  parser.add_option("-n", "--name", action = "store", metavar = "NAME", default = None, \
                    help = "Set the name of the segment to NAME (e.g. DMT-BADMONTH)")
  parser.add_option("-v", "--version", action = "store", metavar = "VERSION", default = None, \
                    help = "Set the numeric version of the segment to VERSION (e.g. 1)")
  parser.add_option("-e", "--explain", action = "store", metavar = "EXPLAIN", default = None, \
                    help = "Set the segment_definer:comment to COMMENT. This should explaining WHAT this flag mean (e.g. \"Light dip 10%\"). Required when --Insert/-I is specified.")
  parser.add_option("-c", "--comment", action = "store", metavar = "COMMENT", default = None, \
                    help = "Set the segment_summary:comment to COMMENT. This should explaining WHY these segments were inserted (e.g. \"Created from hveto results\")")
  parser.add_option("-S", "--summary-file", action = "store", metavar = "FILE", default = None, \
                    help = "Read the segment_summary rows from FILE. This should be a file containing the gps start and end times that the flag was defined (i.e. the union of on and off)")
  parser.add_option("-G", "--segment-file", action = "store", metavar = "FILE", default = None, \
                    help = "Read the segment rows from FILE. This should containin the gps start and end times when the flag was active")
  
  
  return parser.parse_args()
  
    
(options, args) = parse_command_line()
    
# Make sure all necessary command line arguments are properly given
errmsg = ""
if not options.segment_url:
  errmsg += "Error: --segment-url must be specified\n"
if not options.ifos:
  errmsg += "Error: --ifos must be specified\n"
if not options.name:
  errmsg += "Error: --name must be specified\n"
if not options.version:
  errmsg += "Error: --version must be specified\n"
if (not options.explain) and (not options.append):
  errmsg += "Error: --explain must be specified\n"
if not options.comment:
  errmsg += "Error: --comment must be specified\n"
if not options.summary_file:
  errmsg += "Error: --summary-file must be specified\n"
if not options.segment_file:
  errmsg += "Error: --segment-file must be specified\n"
if options.insert and options.append:
  errmsg += "Error: Exactly one of [--insert | --append] can be specified\n"
if (not options.ping) and (not options.insert) and (not options.append):
  errmsg += "One of --ping, --help, --insert, --append must be specified\n"
# --ignore-append-check is too dangerous to allow for direct DB access;
# user must dump to XML, check it, and be authorized to insert via ldbdc
if (not options.output and options.ignore_append_check):
  errmsg += "--ignore-append-check may only be used with --output-file\n"
  
if len(errmsg) and not options.ping:
  print(errmsg, file=sys.stderr)
  print("Run\n    %s --help\nfor more information." % sys.argv[0], file=sys.stderr)
  sys.exit(1)


########################################################################
# open connection to LDBD(W)Server
myClient = segmentdb_utils.setup_database(options.segment_url)

if options.ping:
   try:
     print(myClient.ping())
   except Exception as e:
     print("Error ping --segment-url", str(e))
     sys.exit(1)
else:
########################################################################
  def dtd_uri_callback(uri):
    if uri in ['http://www.ldas-sw.ligo.caltech.edu/doc/ligolwAPI/html/ligolw_dtd.txt',
      'http://ldas-sw.ligo.caltech.edu/doc/ligolwAPI/html/ligolw_dtd.txt']:
      # if the XML file contains a http pointer to the ligolw DTD at CIT then
      # return a local copy to avoid any network problems
      return 'file://localhost' + os.path.join( os.environ["GLUE_PREFIX"],
        'etc/ligolw_dtd.txt' )
    else:
      # otherwise just use the uri in the file
      return uri

  #Create an xml parser, a ligo_lw document parser, and the document
  xmlparser = pyRXP.Parser()
  lwtparser = ldbd.LIGOLwParser()
  segment_md = ldbd.LIGOMetadata(xmlparser,lwtparser)

  ########################################################################
  # Construct local table process, process_params and segment_definer 
  ########################################################################
  # Table process and process_params will be popullated by calling the 
  # process/process_params utility
  
  #myClient = segmentdb_utils.setup_database(options.segment_url)
  doc = ligolw.Document()
  doc.appendChild(ligolw.LIGO_LW())

  ligolwtypes.FromPyType[type(True)] = ligolwtypes.FromPyType[type(0)] 

  process_id = process.register_to_xmldoc(doc, PROGRAM_NAME, options.__dict__,
               version = __version__,
               cvs_entry_time = __date__, comment = options.comment).process_id

  ##########################################################################
  #             Process segment-file and summary-file 
  ##########################################################################
  # create the total time interval and storage for the active segments
  intervals = segments.segmentlist()
  active_segments = segments.segmentlist()

  #-----------------Check version and type existence-------------------------
  if options.insert:
    sql = "SELECT max(version) FROM segment_definer  " 
    sql += "WHERE ifos = '%s' " % options.ifos
    sql += "AND name = '%s' " % options.name
  if options.append:
    sql = "SELECT max(segment_summary.end_time) FROM segment_definer, segment_summary "
    sql += "WHERE segment_summary.segment_def_id = segment_definer.segment_def_id "
    sql += "AND segment_summary.segment_def_cdb=segment_definer.creator_db " 
    sql += "AND segment_definer.ifos = '%s' " % options.ifos
    sql += "AND segment_definer.name = '%s' " % options.name
    sql += "AND segment_definer.version = %d " % int(options.version)
  typexml = myClient.query(sql)
  
  del myClient
  myClient = None


  # parse the result returned from the database query
  type_md = ldbd.LIGOMetadata(xmlparser,lwtparser)
  xmlparser.eoCB = dtd_uri_callback
  type_md.parse(typexml)

  if options.insert: 
    try:
       max_version = type_md.table['segment_definer']['stream'][0][0]
    except IndexError: 
       max_version = 0 # if inserting new flag/version, set max_version to 0
    if (int(options.version) == (max_version + 1)):
       max_end_sum = 0
    else:
       msg = """\nERROR: Wrong Version. Max existing version of %s:%s is: %s. 
Please correct your version number.""" % (options.ifos,options.name,max_version)
       print(msg)
       sys.exit(1)

  if options.append:
     try:
        max_end_sum = type_md.table['segment_definer']['stream'][0][0]
     except IndexError:
        print("\nERROR: segment type %s:%s:%s doesn't exist" % (options.ifos,options.name,options.version))
        sys.exit(1)
 
 
  #-----------------Check summary-file---------------------------------
  # get and check the summary intervals to be inserted
  fh = open(options.summary_file, 'r')
  line_no = 1
  for line in fh.readlines():
    sum_line = line.strip().split(" ")
    if len(sum_line) != 2:
       print("\nERROR: invalid format in your --summary-file row number: %d " % line_no)
       print("Invalid interval is %s" % str(sum_line))
       sys.exit(1)
    current_time = str(gpstime.GpsSecondsFromPyUTC(time.time()))

    # make sure gps time is 9 digits
    #if (len(sum_line[0])!=9 or len(sum_line[1])!=9):
    #   print "\nERROR: gps time must be 9 digits"
    #   print "Invalid interval in your --summary-file row number: %d " % line_no
    #   print "Invalid interval is %s" % str(sum_line)
    #   sys.exit(1)
    ## make sure interval start time is less than and not equal to end time
    if (int(sum_line[0]) >= int(sum_line[1])):
       print("\nERROR: start_time MUST be less than the end_time")
       print("Invalid interval in your --summary-file row number: %d " % line_no)
       print("Invalid interval is %s" % str(sum_line))
       sys.exit(1)
    # make sure interval start time is greater than the max end time in the database
    # unless --ignore-append-check is specified
    elif (int(sum_line[0]) < max_end_sum) and not options.ignore_append_check:
       print("\nERROR: summary start_time MUST be greater than the max summary end_time %d in the database" % max_end_sum)
       print("Invalid interval in your --summary-file row number: %d" % line_no)
       print("Invalid interval is %s" % str(sum_line))
       sys.exit(1)
    # make sure interval end time is less than current time
    elif  (int(sum_line[1]) > int(current_time)):
       print("\nERROR: summary end_time cannot be greater than the current time")
       print("Invalid interval in your --summary-file row number: %d " % line_no)
       print("Invalid interval is %s" % str(sum_line))
       sys.exit(1)
    # if interval is valid, append it to intervals list
    else:
      this_sum = segments.segment(int(sum_line[0]),int(sum_line[1]))
      intervals.append(this_sum)
      line_no += 1
  intervals.coalesce()

  #-----------------Check segment-file---------------------------------
  # get and check the active segments to be inserted
  fh = open(options.segment_file, 'r')
  line_no = 1
  for line in fh.readlines():
    seg_line = line.strip().split(" ")
    if len(seg_line) != 2:
       print("\nERROR: invalid format in your --segment-file row number: %d " % line_no)
       print("Invalid segment is %s" % str(seg_line))
       sys.exit(1)
    # make sure gps time is 9 digits
    #if (len(seg_line[0]) != 9 or len(seg_line[1])!=9): 
    #   print "\nERROR: gps time must be 9 digits"
    #   print "Invalid segment in your --segment-file row number: %d" % line_no
    #   print "Invalid segment is %s" % str(seg_line)
    #   sys.exit(1)
    # make sure segment start_time is less than end_time:
    if int(seg_line[0])>=int(seg_line[1]):
      print("\nERROR: segment start_time MUST be less than the end_time")
      print("Invalid segment in your --segment-file row number: %d" % line_no)
      print("Invalid segment is %s" % str(seg_line))
      sys.exit(1)
    # make sure segment falls in the summary  intervals specified in the --summary-file
    this_seg = segments.segment(int(seg_line[0]),int(seg_line[1])) 
    if this_seg not in intervals:
      print("\nERROR: segment cannot fall outside of the summary intervals specified in your --summary-file")
      print("Invalid segment in your --segment-file row number: %d" % line_no)
      print("Invalid segment is %s" % str(seg_line))
      sys.exit(1)
    # Otherwise, append this segment to the list of active segments
    active_segments.append(this_seg)
    line_no += 1
  active_segments.coalesce()



  #######################################################################################
  #                            Process INSERT
  #######################################################################################
  if options.insert:
    #--------------------------------------------------------------------------#
    # create segment_definer table
    seg_def_table = lsctables.New(lsctables.SegmentDefTable,columns = [
                  'segment_def_id', 'process_id','ifos','name', 'version','comment'])
    doc.childNodes[0].appendChild(seg_def_table)
    segment_definer = lsctables.SegmentDef()

    seg_def_id = seg_def_table.get_next_id()
    segment_definer.segment_def_id = seg_def_id
    segment_definer.process_id = process_id
    segment_definer.ifos = options.ifos
    segment_definer.name = options.name
    segment_definer.version = int(options.version)
    segment_definer.comment = options.explain
    seg_def_table.append(segment_definer)

    #---------------- create local segment table ------------#
    segment_table = lsctables.New(lsctables.SegmentTable, columns = [
                    'segment_id','process_id','segment_def_id','start_time','end_time'])
    doc.childNodes[0].appendChild(segment_table)

    for this_seg in active_segments:
      segment = lsctables.Segment()
      seg_id = segment_table.get_next_id()

      segment.segment_id = seg_id
      segment.process_id = process_id
      segment.segment_def_id = seg_def_id
      segment.start_time = int(this_seg[0])
      segment.end_time = int(this_seg[1])
 
      segment_table.append(segment)

    
    #-------------- create local segment_summary table ------------#
    seg_sum_table = lsctables.New(lsctables.SegmentSumTable, columns = [
                    'segment_sum_id', 'process_id', 'segment_def_id', 'start_time', 'end_time', 'comment'])
    doc.childNodes[0].appendChild(seg_sum_table)

    for this_sum in intervals:
       segment_summary = lsctables.SegmentSum()
       seg_sum_id = seg_sum_table.get_next_id()
       
       segment_summary.segment_sum_id = seg_sum_id
       segment_summary.process_id = process_id
       segment_summary.segment_def_id = seg_def_id
       segment_summary.start_time = this_sum[0]
       segment_summary.end_time = this_sum[1]
       segment_summary.comment = options.comment
   
       seg_sum_table.append(segment_summary)

    fake_file = StringIO.StringIO()
    doc.write(fake_file)

    if options.output:
      fp = open(options.output,"w")
      fp.write(fake_file.getvalue())
      fp.close()
    else:
      myClient = segmentdb_utils.setup_database(options.segment_url)
      myClient.insert(fake_file.getvalue())
     
  #######################################################################################
  #                            Process APPEND
  #######################################################################################
  if options.append:
    #-----------------------------------------------------------# 
    # query the database to get the segment_def_id and creator_db 
    # of this given segment type
    myClient = segmentdb_utils.setup_database(options.segment_url)
    sql = "SELECT creator_db, hex(segment_def_id) AS segment_def_id "
    sql += "From segment_definer WHERE "
    sql += "ifos = '%s' AND " % options.ifos
    sql += "name = '%s' AND " % options.name
    sql += "version = %d " % int(options.version)
    defxml = myClient.query(sql)

    del myClient
    myClient = None

    # parse the result returned from the database query
    seg_def_md = ldbd.LIGOMetadata(xmlparser,lwtparser)
    xmlparser.eoCB = dtd_uri_callback
    seg_def_md.parse(defxml)

    cdb_col = \
     seg_def_md.table['segment_definer']['orderedcol'].index('creator_db')
    sdf_col = \
     seg_def_md.table['segment_definer']['orderedcol'].index('segment_def_id')
    creator_db = seg_def_md.table['segment_definer']['stream'][0][cdb_col]
    seg_def_id = binascii.a2b_hex(seg_def_md.table['segment_definer']['stream'][0][sdf_col])
    #-----------------------------------------------------------------------------#
    # create the elements in the process table that need to be filled in
    process_cols = {
      "process_id" : "ilwd:char",
      "program" : "lstring",
      "version" : "lstring",
      "cvs_repository" : "lstring",
      "cvs_entry_time" : "int_4s",
      "is_online" : "int_4s",
      "node" : "lstring",
      "username" : "lstring",
      "unix_procid" : "int_4s",
      "start_time": "int_4s",
      "end_time": "int_4s",
      "ifos" : "lstring",
      "comment" : "lstring",
     }

    process_ocols = [
     "process_id",
     "program",
     "version",
     "cvs_repository",
     "cvs_entry_time",
     "is_online",
     "node",
     "username",
     "unix_procid",
     "start_time",
     "end_time",
     "ifos",
     "comment"
    ]

    process_stream = [(
     "process:process_id:0",
     os.path.basename(sys.argv[0]), 
     __version__, 
     __src__,
     gpstime.GpsSecondsFromPyUTC( time.mktime( time.strptime( 
      __date__, "%Y-%m-%d %H:%M:%S +0000" ) ) ),
     0,
     socket.gethostname(),
     pwd.getpwuid(os.geteuid())[0],
     os.getpid(),
     gpstime.GpsSecondsFromPyUTC(time.time()),
     gpstime.GpsSecondsFromPyUTC(time.time()),
     options.ifos,
     options.comment
     )]

    process_dict = { 
    'pos' : 0,
    'column' : process_cols,
    'orderedcol' : process_ocols,
    'stream' : process_stream
    }

    #----------- create local process_params table ------#
    process_params_cols = {
    "process_id" : "ilwd:char",
    "program" : "lstring",
    "type" : "lstring",
    "value" : "lstring",
    "param":"lstring"
    }
    
    process_params_ocols = [
    "process_id",
    "program",
    "type",
    "value",
    "param"
    ]

    process_params_stream = []
    for key,val in six.iteritems(vars(options)): 
      process_params_stream.append(
      tuple([   
      "process:process_id:0",
      os.path.basename(sys.argv[0]), 
      "lstring",
      str(val),
      "--" + str(key).replace('_','-')
    ])) 

    process_params_dict = {
    'pos' : 0,
    'column' : process_params_cols,
    'orderedcol' : process_params_ocols,
    'stream' : process_params_stream
    }

    #------------ create local segment table ------------#
    segment_cols = {
      "process_id" : "ilwd:char",
      "segment_def_id" : "ilwd:char_u",
      "segment_def_cdb" : "int_4s",
      "segment_id" : "ilwd:char",
      "start_time" : "int_4s",
      "end_time" : "int_4s",
    }

    segment_ocols = [
     "process_id",
     "segment_def_id",
     "segment_def_cdb",
     "segment_id",
     "start_time",
     "end_time",
    ]

    segment_stream = []
    seg_id = 0
    for this_seg in active_segments:
      segment_stream.append (
      tuple(["process:process_id:0",
      seg_def_id,
      creator_db,
      "segment:segment_id:" + str(seg_id),
      int(this_seg[0]),
      int(this_seg[1])
      ]))
      seg_id += 1
 
    segment_dict = { 
     'pos' : 0,
     'column' : segment_cols,
     'orderedcol' : segment_ocols,
     'stream' : segment_stream
    } 

    #------------ create local segment_summary table  ------------#
    # create the elements in the segment_summary table that need to be filled in
    segment_summary_cols = {
    "process_id" : "ilwd:char",
    "segment_def_id" : "ilwd:char_u",
    "segment_def_cdb" : "int_4s",
    "segment_sum_id" : "ilwd:char",
    "start_time" : "int_4s",
    "end_time" : "int_4s",
    "comment" : "lstring"
     }

    segment_summary_ocols = [
    "process_id",
    "segment_def_id",
    "segment_def_cdb",
    "segment_sum_id",
    "start_time",
    "end_time",
    "comment"
    ]
 
    segment_summary_stream = []
    sum_id = 0
    for this_sum in intervals:
      segment_summary_stream.append(
      tuple([
      "process:process_id:0",
      seg_def_id,
      creator_db,
      "segment_summary:segment_sum_id" + str(sum_id),
      int(this_sum[0]),
      int(this_sum[1]),
      options.comment
      ]))
      sum_id += 1
    
 
    segment_summary_dict = { 
    'pos' : 0,
    'column' : segment_summary_cols,
    'orderedcol' : segment_summary_ocols,
    'stream' : segment_summary_stream
    }
    
    #append all the tales we have created for APPEND
    segment_md.table['process']=process_dict
    segment_md.table['process_params']=process_params_dict
    segment_md.table['segment']=segment_dict
    segment_md.table['segment_summary']=segment_summary_dict


    if options.output:
      fp = open(options.output,"w")
      fp.write(segment_md.xml())
      fp.close()
    else:
      myClient = segmentdb_utils.setup_database(options.segment_url)
      myClient.insert(segment_md.xml())

sys.exit(0)
 




#
#
# Copyright (C) 2009  Larne Pekowsky
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


#
# =============================================================================
#
#                                   Preamble
#
# =============================================================================
#


"""
Takes one or more files with sngl_$x triggers (where x = inspiral or burst or...)
and a veto files, and filters or passes all triggers in vetoed times
"""


from __future__ import print_function
from optparse import OptionParser
import sqlite3
import sys
import os
import pwd
import tempfile
import re
import time
import datetime

from glue import gpstime
from glue.ligolw import ligolw
from glue.ligolw import lsctables
from glue.ligolw import utils
from glue.ligolw.utils import ligolw_sqlite
from glue.ligolw import dbtables

from glue.segmentdb import segmentdb_utils

from glue.ligolw.utils import process

import glue.ligolw.types as ligolwtypes

PROGRAM_NAME = sys.argv[0].replace('./','')
PROGRAM_PID  = os.getpid()
try:
        USER_NAME = os.getlogin()
except:
        USER_NAME = pwd.getpwuid(os.getuid())[0]

from glue import git_version
__author__ = "Larne Pekowsky <lppekows@physics.syr.edu>"
__date__ = git_version.date
__version__ = git_version.id


#
# =============================================================================
#
#                                 Command Line
#
# =============================================================================
#


def parse_command_line():
    """
    Parse the command line, return an options object
    """

    def require(flag, message):
        if not flag:
            print(message, file=sys.stdout)
            sys.exit(1)


    parser = OptionParser(
        version = "%prog CVS $Header$",
        usage   = "%prog -v|--veto-file filename [options] trigger_file1 trigger_file2 ...",
        description = "Reads one or more segment files and a veto file and generates files of veto segments"        
	)
    
    parser.add_option("-r", "--trigger-dir",  metavar = "trigger_dir",  help = "Top-level directory with trigger files")
    parser.add_option("-t", "--type",         metavar = "type",         help = "type of trigger [burst|inspiral]")
    parser.add_option("-i","--ifos",          metavar = "ifos",         help = "IFOs to process")
    parser.add_option("-c","--categories",    metavar="categories",     help = "Categories to process")

    parser.add_option("-n", "--include-veto",  action="store_true",     help = "Only return triggers that are inside vetoed times") 
    parser.add_option("-x", "--exclude-veto",  action="store_true",     help = "Only return triggers that are outside vetoed times")
    parser.add_option("-o", "--output-dir",    metavar = "output_dir",  help = "Directory to write output XML (default cwd)")
    parser.add_option("-q", "--dq-dir",        metavar = "dq_dir",      help = "Top-level directory with DQ files")

    parser.add_option("-s", "--gps-start-time",  metavar="gps_start_time", help = "Start time for triggers")
    parser.add_option("-e", "--gps-end-time",    metavar="gps_end_time",   help = "End time for triggers")

    parser.add_option("-p", "--separate",        metavar="separate",    action="store_true", help="Do separate categories")
    parser.add_option("-m", "--cumulative",      metavar="cumulative",  action="store_true",  help="Do cumulative categories")


    parser.add_option("-v","--veto-file",      metavar="veto_file",      help = "Veto definer file")
    parser.add_option("-f","--timestamp-file", metavar="timestamp_file", help = "File holding gps start time")

    options, others = parser.parse_args()

    require(options.veto_file, "missing required argument --veto-dir")
    require(options.trigger_dir, "missing required argument --trigger-dir")
    require(options.dq_dir, "missing required argument --dq-dir")
    require(options.type, "missing required argument --type [burst|inspiral]")
    require(options.type in ['burst', 'inspiral'], "type must be one of [burst|inspiral]")
    require(options.include_veto or options.exclude_veto, "missing required argument, either --include-veto or --exclude-veto")

    require(options.separate or options.cumulative, "missing required argument, at least one of --separate or --cumulative")

    require(options.timestamp_file or (options.gps_start_time and options.gps_end_time), 
            "missing required argument, either --timestamp-file or  --gps-start-time and --gps-end-time")

    return options



def get_output_name(ifo, category, gps_start_time, gps_end_time, output_dir, isCumulative):
    first_four = gps_start_time / 100000

    full_dir_path = '%s/%s_CAT%d%s/%s-MBTA-%d' % (output_dir, ifo, category, isCumulative and '_CUMULATIVE' or '', ifo, first_four)

    try:
        os.makedirs(full_dir_path)
    except:
        pass

    return '%s/%s-MBTA_CAT%d-%d-%d.xml' % (full_dir_path, ifo, category, gps_start_time, (gps_end_time - gps_start_time))



#
# =============================================================================
#
#                                     Main
#
# =============================================================================
#

if __name__ == '__main__':
    # Fd2XMLIOServer populates the event_id early, but ihope leaves 
    # it as 0 initially.  Turnning off the unique constraint allows
    # us to load both.
    del lsctables.SnglInspiralTable.constraints
    del lsctables.SnglInspiralTable.next_id

    options = parse_command_line()    

    # Run up to now or the user-provided time
    if options.gps_end_time:
        gps_end_time = int(options.gps_end_time)
    else:
        gps_end_time = gpstime.GpsSecondsFromPyUTC(time.time())

    # Find the last time we ran (in principle we could get this from the time
    # stamp on the HTML file, but we should allow for the possibility that 
    # someone might hand-edit that file)
    if options.gps_start_time:
        gps_start_time = int(options.gps_start_time)
    elif os.path.exists(options.timestamp_file):
        f = open(options.timestamp_file)
        gps_start_time = int(next(f))
        f.close()
    else:
        # Looks like we've never run before, start a day ago
        gps_start_time = gps_end_time - (60 * 60 * 24)

    # Find trigger files
    trigger_files = segmentdb_utils.get_all_files_in_range(options.trigger_dir, gps_start_time, gps_end_time, pad=0)
    
    # If there are no files in the time range then there's nothing to do
    if not trigger_files:
        # We don't update the timestamp here.  If we did we would never
        # give data coming in with some latency the chance to catch up.
        sys.exit(0)

    # Due to latency issues the last available time may not be the last 
    # time we asked for.  Find the real end time from the file names
    trigger_files.sort()

    m = re.match('.*-([0-9]+)-([0-9]+).xml',trigger_files[0])
    first_trigger_time = int(m.group(1)) + int(m.group(2))

    m = re.match('.*-([0-9]+)-([0-9]+).xml',trigger_files[-1])
    last_trigger_time = int(m.group(1)) + int(m.group(2))

    # and dq files - we don't need files outside the range of time for
    # which we have triggers
    dq_files = segmentdb_utils.get_all_files_in_range(options.dq_dir, first_trigger_time, last_trigger_time)

    # Again, if there are no files there's nothing to do
    if not dq_files:
        sys.exit(0)

    # Sort them and look at the last available time
    dq_files.sort()

    m  = re.match('.*-([0-9]+)-([0-9]+).xml',dq_files[-1])
    last_dq_time = int(m.group(1)) + int(m.group(2))

    # We should only run up to the earlier of the times available
    # to ensure we have both DQ and triggers
    gps_end_time = min(gps_end_time, last_trigger_time, last_dq_time)

    # If we're looking at less than two seconds of time it's not worth it...
    if gps_end_time - gps_start_time < 2:
        sys.exit(0)

    # Build the full list of XML files to load, includiing the veto_definer file
    xml_files = trigger_files + [options.veto_file] + dq_files


    # Load the files into sqlite. 
    handle, temp_db = tempfile.mkstemp(suffix='.sqlite')
    os.close(handle)
    
    target          = dbtables.get_connection_filename(temp_db, None, True, False)
    connection      = ligolw_sqlite.setup(target)
    cursor          = connection.cursor()

    ligolw_sqlite.insert_from_urls(connection, xml_files)

    table_name = 'sngl_' + options.type

    # Each trigger file and DQ file etc has it's own process and
    # process param table, which doesn't really give us anything
    # useful and has a lot of overhead.
    cursor.execute("DELETE FROM process")
    cursor.execute("DELETE FROM process_params")

    cursor.execute("INSERT INTO process(process_id, program, ifos) VALUES('process:process_id:0', 'ligolw_veto_sngl_triggers', '%s')" % options.ifos)


    # Make sure the table exists
    theClass  = lsctables.TableByName[table_name]
    statement = "CREATE TABLE IF NOT EXISTS " + table_name + " (" + ", ".join(["%s %s" % (key, ligolwtypes.ToSQLiteType[theClass.validcolumns[key]]) for key in theClass.validcolumns]) + ")"

    cursor.execute(statement)

    # The trigger files may not lie exactly on the start/end
    # boundaries we specified, delete anything outside this range
    cursor.execute('DELETE FROM %s WHERE end_time < %d OR end_time >= %d' % (table_name, gps_start_time, gps_end_time))

    # Clean up process_id references
    cursor.execute("UPDATE %s SET process_id='process:process_id:0'" % table_name)
    cursor.execute("UPDATE search_summary SET process_id='process:process_id:0'")

    # Determine which tables we're outputting
    output_tables = []
    for name in ['process','process_params',table_name,'search_summary','summ_value']:
        count = int(cursor.execute("select count(*) from sqlite_master where name='%s'" % name).fetchone()[0])
        if count:
            output_tables.append(name)

    # Write out a set of triggers that cuts out non-science time
    for ifo in options.ifos.split(','):
        subquery  = "SELECT %s.end_time FROM %s, segment, segment_definer " % (table_name, table_name)
        subquery += "   WHERE segment_definer.segment_def_id = segment.segment_def_id "
        subquery += "     AND segment_definer.name = 'DMT-SCIENCE' "
        subquery += "     AND segment_definer.ifos = '%s' " % (ifo)
        subquery += "     AND (%s.end_time BETWEEN segment.start_time AND segment.end_time)" % (table_name)

        sql       = 'DELETE FROM %s WHERE end_time NOT IN (%s)' % (table_name, subquery)

        cursor.execute(sql)

        base_dir   = options.output_dir or '.'
        first_four = gps_start_time / 100000
        outdir     = "%s/%s-SCI/%s-SCI-%d" % (base_dir, ifo, ifo, first_four)
        outname    = "%s/%s-SCI-%d-%d.xml" % (outdir, ifo, gps_start_time, (gps_end_time - gps_start_time))

        try:
            os.makedirs(outdir)
        except:
            pass

        ligolw_sqlite.extract(connection, outname, table_names=output_tables)


    # Backup the sngl_ table
    sql = cursor.execute("SELECT sql FROM sqlite_master WHERE name='%s'" % table_name).fetchone()[0]
    pos = sql.index(table_name) + len(table_name)
    sql = sql[0:pos] + '_backup' + sql[pos:]

    cursor.execute(sql)
    cursor.execute('INSERT INTO %s_backup SELECT * FROM %s' % (table_name, table_name))



    ops = []
    if options.separate:
        ops.append(' = ')
    if options.cumulative:
        ops.append(' <= ')

    for op in ops:
        for ifo in options.ifos.split(','):
            for category in map(int, options.categories.split(',')):
                cursor.execute('DELETE FROM %s' % table_name)
                cursor.execute("INSERT INTO %s SELECT * FROM %s_backup WHERE ifo='%s'" % (table_name, table_name, ifo))

                # Delete any triggers whose end time falls within a vetoed segment
                subquery  = "SELECT %s.end_time FROM %s, segment, segment_definer, veto_definer " % (table_name, table_name)
                subquery += "   WHERE segment_definer.segment_def_id = segment.segment_def_id "
                subquery += "     AND segment_definer.name = veto_definer.name "
                subquery += "     AND segment_definer.ifos = veto_definer.ifo "
                subquery += "     AND segment_definer.version = veto_definer.version "
                subquery += "     AND veto_definer.category %s %d " % (op, category)
                subquery += "     AND veto_definer.ifo = '%s'"      % (ifo)
                subquery += "     AND %s.end_time BETWEEN segment.start_time AND segment.end_time" % (table_name)


                sql =  'DELETE FROM ' + table_name
                sql += ' WHERE end_time %s IN ' % (options.exclude_veto and ' ' or 'NOT')
                sql += ' (%s)' % subquery

                cursor.execute(sql)

                # Write out
                base_dir   = options.output_dir or '.'
                ext        = op == ' <= ' and '_CUMULATIVE' or ''
                first_four = gps_start_time / 100000
                outdir     = "%s/%s-CAT%d%s/%s-CAT%d%s-%d" % (base_dir, ifo, category, ext, ifo, category, ext, first_four)

                try:
                    os.makedirs(outdir)
                except:
                    pass

                outname = "%s/%s-CAT%d%s-%d-%d.xml" % (outdir, ifo, category, ext, gps_start_time, (gps_end_time - gps_start_time))

                ligolw_sqlite.extract(connection, outname, table_names=output_tables)


    # If we started from a timestamp file, update it
    if options.timestamp_file:
        f = open(options.timestamp_file, 'w')
        print((gps_end_time + 1), file=f)
        f.close()

    # Clean up
    os.remove(temp_db)




#
# Copyright (C) 2010  Peter Couvares  <pfcouvar@syr.edu>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""
This intended as a drop-in replacement for ligolw_segments_from_cats
which splits the specified query into multiple shorter queries (by GPS
time) and recombines the results.  It supports a single additional
optional argument, --max_gps_duration, to specify the maximum gps time
of each constituent query (default is one week).
"""

from __future__ import print_function
import sys
import os
import pwd
import tempfile
import re
from optparse import OptionParser

import ligo.segments
from ligo.segments.utils import segmentlist_range
from glue import git_version

PROGRAM_NAME = sys.argv[0].replace('./','')
PROGRAM_PID  = os.getpid()
try:
        USER_NAME = os.getlogin()
except:
        USER_NAME = pwd.getpwuid(os.getuid())[0]


__author__ = "Peter Couvares <pfcouvar@syr.edu>"
__version__ = "git id %s" % git_version.id
__date__ = git_version.date

# lifted from ligolw_segments_from_cats, with --max-gps-duration added
def parse_command_line():
    """
    Parse the command line, return an options object
    """

    parser = OptionParser(
        version = "Name: %%prog\n%s" % git_version.verbose_msg,
        usage   = "%prog -v|--veto-file filename [options]",
        description = "Reads one or more segment files and a veto file and generates files of veto segments"        
	)
    
    parser.add_option("-v", "--veto-file",    metavar = "veto_file",    help = "veto XML file (required).")
    parser.add_option("-o", "--output-dir",   metavar = "output_dir",   default = '.',          help = "Directory to write output (default=cwd).")
    parser.add_option("-k", "--keep-db",      metavar = "keep_db",      action  = "store_true", help = "Keep sqlite database.")
                      
    parser.add_option("-t", "--segment-url",  metavar = "segment_url", help = "Segment URL")
    parser.add_option("-d", "--database",     metavar = "use_database", action = "store_true", help = "use database specified by environment variable S6_SEGMENT_SERVER")
    parser.add_option("-f", "--dmt-file",     metavar = "use_files", action = "store_true", help = "use files in directory specified by environment variable ONLINEDQ")
    parser.add_option("-c", "--cumulative-categories",   action = "store_true", help = "If set the category N files will contain all segments in categories <= N")
    parser.add_option("-p", "--separate-categories",     action = "store_true", help = "If set the category N files will contain only category N")
    
    parser.add_option("-s", "--gps-start-time", metavar = "gps_start_time", help = "Start of GPS time range")
    parser.add_option("-e", "--gps-end-time",   metavar = "gps_end_time", help = "End of GPS time range")

    parser.add_option("--max-gps-duration",   metavar = "max_gps_duration", help = "Maximum duration of an individual query; larger durations will be split into multiple queries and recombined.")

    options, others = parser.parse_args()

    if not options.gps_start_time:
        raise ValueError( "missing required argument --gps-start-time" )
    
    if not options.gps_end_time:
        raise ValueError( "missing required argument --gps-end-time" )

    return options


# generates a modified version of the provided argv array, minus a
# given argument (and optionally its value)
def strip_arg(argv, long_name, short_name=None, remove_value=True):
    argn = 0
    while argn < len(argv):
        arg = argv[argn]

	# if the argument includes a value delimited by '=', separate it
        argsplit = arg.split('=')
        arg = argsplit[0]
	
        if arg == long_name or arg == short_name:
            # if we are to remove the associated value, and that value
            # is not already inside the arg (via '='), then skip the
            # subsequent argument
            if remove_value and len(argsplit)==1:
                argn += 1
        else:
            yield argv[argn]
        argn += 1


if __name__ == '__main__':
    options = parse_command_line()

    if options.max_gps_duration is None:
        max_gps_duration = 60*60*24*7
    else:
        max_gps_duration = int(options.max_gps_duration)

    gps_start_time = int(options.gps_start_time)
    gps_end_time = int(options.gps_end_time)
    gps_duration = gps_end_time - gps_start_time

    if max_gps_duration > gps_duration:
        max_gps_duration = gps_duration

#    print gps_start_time
#    print gps_end_time

    gps_segmentlist = list(segmentlist_range(gps_start_time, gps_end_time, max_gps_duration))
    if gps_segmentlist[-1][1] != int(gps_end_time):
        gps_segmentlist.append(ligo.segments.segment(gps_segmentlist[-1][1],gps_end_time))

    print("segmentlist =", gps_segmentlist)

    argv = sys.argv[1:]
#    print "original argv =", argv
    argv = list(strip_arg(argv, "-s", "--gps-start-time"))
    argv = list(strip_arg(argv, "-e", "--gps-end-time"))
    argv = list(strip_arg(argv, "-o", "--output-dir"))
    argv = list(strip_arg(argv, "--max-gps-duration"))

#    print "stripped argv =", argv

    temp_dirs = []

    for segment in gps_segmentlist:
        temp_dir = tempfile.mkdtemp()
        temp_dirs.append(temp_dir)
        temp_argv = list(argv)
        temp_argv.append("-s %s" % segment[0])
        temp_argv.append("-e %s" % segment[1])
        temp_argv.append("-o %s" % temp_dir)
#        print "temp_argv =", temp_argv
        os.system("ligolw_segments_from_cats %s" % ' '.join(temp_argv))

    # in theory, the same output files should be present in each output
    # dir, so we just need to look in one to figure out the names
    # e.g., "L1-VETOTIME_CAT2-970271943-259200.xml"

    prefix = re.compile(r"^[A-Z][0-9]-VETOTIME_CAT[0-9]")
    temp_files = [prefix.match(s).group() for s in os.listdir(temp_dirs[0])]

    for file in temp_files:
        file_instances = ["%s/%s-*-*.xml" % (dir, file) for dir in temp_dirs]
        os.chdir(options.output_dir)
#        os.system("pwd")
        result_file = "%s-%d-%d.xml" % (file, gps_start_time, gps_duration)
        os.system("ligolw_add --output %s %s" % (result_file, ' '.join(file_instances)))
#        os.system("rm -v %s" % ' '.join(file_instances))

#    for dir in temp_dirs:
#        os.rmdir(dir)



#
#
# Copyright (C) 2014  Alex Nitz
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""
Takes one or more XML files containing segment lists and creates a new
segment list that represents the intersection.
"""
from __future__ import print_function
import copy
from optparse import OptionParser
from ligo.segments import segmentlist
from glue.ligolw import ligolw, lsctables, utils
from glue.ligolw.utils import process, segments, ligolw_add
from glue import git_version

class ContentHandler(ligolw.LIGOLWContentHandler):
    pass
lsctables.use_in(ContentHandler)

def parse_command_line():
    """
    Parse the command line, return an options object
    """

    parser = OptionParser(
        version = "Name: %%prog\n%s" % git_version.verbose_msg,
        description = "Reads one or more files containing veto segments and "
                      "generates a file with the cumulative segments."        
	)
    
    parser.add_option("--output", help = "The output name of XML file "
                                         "containing the cumulative segments.")
    parser.add_option("--segment-name", help = "The name of the cumulative  "
                                               "segments.")
    options, args = parser.parse_args()
    
    for opt in ['segment-name', 'output']:
        opt = opt.replace('-', '_')
        if not opt in vars(options):
            raise parser.error(' %s is required' % opt)
    return options, args
    
if __name__ == '__main__':
    opt, arg = parse_command_line()

    # start a new document
    doc = ligolw.Document()
    doc.appendChild(ligolw.LIGO_LW())
    lwdoc = doc.childNodes[0]
      
    # add the original documents together and pull out the combined process
    # table #FIXME do I really need to propagate this?
    try:
        tmp_doc = ligolw.Document()
        ligolw_add.ligolw_add(tmp_doc, arg)
        proc_table = lsctables.ProcessTable.get_table(tmp_doc)
        param_table = lsctables.ProcessParamsTable.get_table(tmp_doc)
        lwdoc.appendChild(proc_table)
        lwdoc.appendChild(param_table)
        veto_table = lsctables.VetoDefTable.get_table(tmp_doc)
        lwdoc.appendChild(veto_table)
    except:
        # This is a *serious* error, should fail here, not skip
        print('No process table found in input files, ignoring.....')

    # Register ourselves
    proc_id = process.register_to_xmldoc(doc, 'ligolw_combine_segments', 
                               opt.__dict__, version=git_version.id).process_id

    # Read in the given xml files containing segments
    ifos = set()
    valid = None
    segment_list = segmentlist([])
    for filename in arg:
        indoc = utils.load_filename(filename, 
                             contenthandler=ContentHandler)
        segs = segments.LigolwSegments(indoc)
        for seglist in segs:
            # valid should be the intersection of the valid portion
            # in all files
            if valid is None:
                valid = seglist.valid
            else:
                valid = valid and seglist.valid
            
            # the union of the valid segments in all files
            segment_list += seglist.active   
            ifos.update(seglist.instruments)
            segment_list, seglist.active

    # Create the segment definer table
    seg_def = lsctables.New(lsctables.SegmentDefTable)
    def_row = lsctables.SegmentDef()
    def_row.ifos = lsctables.ifos_from_instrument_set(ifos)
    def_row.name = opt.segment_name
    def_row.process_id = proc_id
    def_row.segment_def_id = seg_def.get_next_id()
    def_row.version = 1
    seg_def.append(def_row)
    lwdoc.appendChild(seg_def)

    # Create the segment summary table
    seg_sum = lsctables.New(lsctables.SegmentSumTable)
    sum_row = lsctables.SegmentSum()
    sum_row.start_time = valid[0][0]
    sum_row.start_time_ns = 0
    sum_row.end_time = valid[0][1]
    sum_row.end_time_ns = 0
    sum_row.process_id = proc_id
    sum_row.segment_def_id = def_row.segment_def_id
    sum_row.segment_sum_id = seg_sum.get_next_id()
    seg_sum.append(sum_row)
    lwdoc.appendChild(seg_sum)
    
    # These are filled with default (meaningless) values
    def_row.comment = ''
    sum_row.comment = ''
    def_row.creator_db = 0
    sum_row.creator_db = 0
    def_row.insertion_time = 0
    sum_row.segment_def_cdb = 0
    
    # Create the segment table
    seg_table = lsctables.New(lsctables.SegmentTable)
    base_row = lsctables.Segment()
    base_row.creator_db = 0
    base_row.segment_def_cdb = 0 
    base_row.process_id = proc_id  
    base_row.segment_def_id = def_row.segment_def_id
    for seg in segment_list:
        row = copy.copy(base_row)
        row.set(seg)
        row.segment_id = seg_table.get_next_id()
        seg_table.append(row)
    lwdoc.appendChild(seg_table)      

    # FIXME: Segment utils currently writes tables in non-standard format.
    #        until this is fixed, we need some additional hacking to make files
    #        compatible.

    # Need to add values to process table
    values = {}
    values["process"] = {}
    values["process"]["domain"] = None
    values["process"]["jobid"] = None
    values["process"]["is_online"] = False

    processTab = lsctables.ProcessTable.get_table(doc)
    for column_name in values["process"]:
        try:
            column = processTab.appendColumn(column_name)
        except ValueError:
            # Column already present
            continue
        # Set values
        for i in range(len(processTab)):
            column[i] = values["process"][column_name]
    
    utils.write_filename(doc, opt.output, gz=opt.output.endswith('gz'))




#
#
# Copyright (C) 2009  Larne Pekowsky, Ping Wei
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


#
# =============================================================================
#
#                                   Preamble
#
# =============================================================================
#


"""
This provides the means to answer several questions posed against either the
segment database or a collection of DMT XML files:

  * What DQ flags exist in the database? ligolw_segment_query --show-types
  * When was a given DQ flag defined? ligolw_segment_query --query-types 
  * When was a given flag active? ligolw_segment_query --query-segments
"""


from __future__ import print_function
from optparse import OptionParser
import sqlite3
import sys
import os
import operator
import pwd
import glob
import time
import socket
import tempfile
import urllib

import ligo.segments

from glue.ligolw import ligolw
from glue.ligolw import table
from glue.ligolw import lsctables
from glue.ligolw import utils

from glue.ligolw.utils import ligolw_add
from glue.ligolw.utils import process
from glue.segmentdb import query_engine
from glue.segmentdb import segmentdb_utils

from glue.ligolw.utils import ligolw_sqlite
from glue.ligolw import dbtables

from glue import git_version
import six

PROGRAM_NAME = sys.argv[0].replace('./','')
PROGRAM_PID  = os.getpid()
try:
        USER_NAME = os.getlogin()
except:
        USER_NAME = pwd.getpwuid(os.getuid())[0]


__author__  = "Larne Pekowsky <lppekows@physics.syr.edu>, Ping Wei <piwei@syr.edu>"
__version__ = "git id %s" % git_version.id
__date__ = git_version.date


#
# =============================================================================
#
#                                 Command Line
#
# =============================================================================
#


def parse_command_line():
    """
    Parse the command line, return an options object
    """

    parser = OptionParser(
        version = "Name: %%prog\n%s" % git_version.verbose_msg,
        usage       = "%prog [ --version | --ping | --show-types | --query-types | --query-segments ]  [ --segment | --database | --dmt-files ] options ",
        description = "Performs a number of queries against either a set of DMT files or a segment database"
	)


    # Major modes
    parser.add_option("-p", "--ping",           action = "store_true", help = "Ping the target server")
    parser.add_option("-y", "--show-types",     action = "store_true", help = "Returns a xml table containing segment type information: ifos, name, version, segment_definer.comment, segment_summary.start_time, segment_summary.end_time, segment_summary.comment")
    parser.add_option("-u", "--query-types",    action = "store_true", help = "Returns a ligolw document whose segment_definer table includes all segment types defined in the given period and included by include-segments and whose segment_summary table indicates the times for which those segments are defined.")
    parser.add_option("-q", "--query-segments", action = "store_true", help = "Returns a ligolw document whose segment table contains the times included by the include-segments flag and excluded by exclude-segments")

    # Time options
    parser.add_option("-s", "--gps-start-time", metavar = "gps_start_time", help = "Start of GPS time range")
    parser.add_option("-e", "--gps-end-time",   metavar = "gps_end_time", help = "End of GPS time range")


    # Data location options
    parser.add_option("-t", "--segment-url",    metavar = "segment_url", help = "Segment URL. Users have to specify either 'https://' for a secure connection or 'http://' for an insecure connection in the segment database url. For example, '--segment-url=https://segdb.ligo.caltech.edu'. No need to specify port number. ")
    parser.add_option("-d", "--database",   metavar = "use_database", action = "store_true", help = "use database specified by environment variable S6_SEGMENT_SERVER. For example, 'S6_SEGMENT_SERVER=https://segdb.ligo.caltech.edu'")
    parser.add_option("-f", "--dmt-files",   metavar = "use_files", action = "store_true", help = "use files in directory specified by environment variable ONLINEDQ, for example, 'ONLINEDQ=file:///path_to_dmt'. 'file://' is the prefix, the acutal directory to DMT xml files starts with '/'.")


    # Other options
    parser.add_option("-a", "--include-segments", metavar = "include_segments", help = "This option expects a comma separated list of a colon separated sublist of interferometer, segment type, and version. The union of segments from all types and versions specified is returned. Use --show-types to see what types are available.   For example: --include-segment-types H1:DMT-SCIENCE:1,H1:DMT-INJECTION:2 will return the segments for which H1 is in either SCIENCE version 1 or INJECTION version 2 mode. If version information is not provided, the union of the segments of the latest version of requested segment type(s) will be returned.")

    parser.add_option("-b", "--exclude-segments", metavar = "exclude_segments", help = "This option has to be used in conjunction with --include-segment-types --exclude-segment-types subtracts the union of unwanted segments from the specified types from the results of --include-segment-types. If version information is not provided, --exclude-segment-types subtracts the union of segments from the latest version of the specified segment types. For example, --include-segment-types H1:DMT-SCIENCE:1,H1:DMT-INJECTION:2 --exclude-segment-types H1:DMT-WIND:1,H1:DMT-NOT_LOCKED:2,H2:DMT-NOT_LOCKED:2 will subtract the union of segments which H1 is in version 1 WIND and H1,H2 is version 2 NOT_LOCKED from the result of --include-segment-types H1:DMT-SCIENCE:1,H1:DMT-INJECTION:2")


    parser.add_option("-S", "--strict-off", metavar = "use_strict", action = "store_true", help = "The default behavior is to truncate segments so that returned segments are entirely in the interval [gps-start-time, gps-end-time).  However if this option is given, the entire non-truncated segment is returned if any part of it overlaps the interval.")

    parser.add_option("-n", "--result-name", metavar = "result_name", default = "RESULT", help = "Name for result segment definer (default = RESULT)") 

    parser.add_option("-o", "--output-file",   metavar = "output_file", help = "File to which output should be written.  Defaults to stdout.")
    
    options, others = parser.parse_args()

    # Make sure we have exactly one thing to do
    count = 0
    for arg in [options.ping, options.query_types, options.show_types, options.query_segments]:
        if arg:
            count += 1
            
    if count != 1:
        raise ValueError("Exactly one of [ --ping | --show-types | --query-types | --query-segments ] must be provided")
    
    
    # Make sure we have required arguments
    database_location = None
    file_location     = None

    # Make sure we know who to contact for data
    if options.segment_url:
        if options.segment_url.startswith('ldbd') or options.segment_url.startswith('http'):
            database_location = options.segment_url
        elif options.segment_url.startswith('file:'):
            file_location = options.segment_url[len('file://'):]
        else:
            tmp_dir = tempfile.mkdtemp()
        
            # Grab the part of the name after the last slash
            pos     = options.segment_url[::-1].find('/')
            fname   = (pos > -1) and options.segment_url[ -1 * pos:] or "dmt.xml" 

            inurl   = urllib.urlopen(options.segment_url)
            outfile = open(tmp_dir + "/" + fname, 'w')
            for l in inurl:
                print(l, end=' ', file=outfile)

            inurl.close()
            outfile.close()
            file_location = tmp_dir
    elif options.database:
        if 'S6_SEGMENT_SERVER' not in os.environ:
            raise ValueError( "--database specified but S6_SEGMENT_SERVER not set" )
        database_location = os.environ['S6_SEGMENT_SERVER']
    elif options.dmt_files:
        if 'ONLINEDQ' not in os.environ:
            raise ValueError( "--dmt-files specified but ONLINEDQ not set" )
        tmp = os.environ['ONLINEDQ']
        if tmp.startswith('file://'):
            tmp = tmp[len('file://'):]
        file_location = tmp
    else:
        raise ValueError( "One of [ --segment-url | --database | --dmt-files ] must be provided" )
        

    # Unless we're pinging, make sure we have start and end times
    if options.ping:
        if not database_location:
            raise ValueError("--ping requires [ --segment-url https://... | --database ]")
    else:
        if not options.gps_start_time:
            raise ValueError( "missing required argument --gps-start-time" )
    
        if not options.gps_end_time:
            raise ValueError( "missing required argument --gps-end-time" )

        if not options.show_types and not options.include_segments:
            raise ValueError( "missing required argument --include-segments")
    
    return options, database_location, file_location




#
# =============================================================================
#
#                                 General utilities
#
# =============================================================================
#


def seg_spec_to_sql(spec):
    """Given a string of the form ifo:name:version, ifo:name:* or ifo:name
    constructs a SQL caluse to restrict a search to that segment definer"""

    parts = spec.split(':')
    sql   = "(segment_definer.ifos = '%s'" % parts[0]

    if len(parts) > 1 and parts[1] != '*':
        sql += " AND segment_definer.name = '%s'" % parts[1]
        if len(parts) > 2 and parts[2] != '*':
            sql += " AND segment_definer.version = %s" % parts[2]

    sql += ')'

    return sql



#
# The results of show-types is a join against segment_definer and segment
# summary, and so does not fit into an existing table type.  So here we
# define a new type so that the ligolw routines can generate the XML
#
class ShowTypesResultTable(table.Table):
    tableName = "show_types_result"

    validcolumns = {
        "ifos": "lstring",
        "name": "lstring",
        "version": "int_4s",
        "segment_definer_comment": "lstring",
        "segment_summary_start_time": "int_4s",
        "segment_summary_end_time": "int_4s",
        "segment_summary_comment": "lstring"
        }
    


class ShowTypesResult(object):
    __slots__ = tuple(ShowTypesResultTable.validcolumns.keys())

    def get_pyvalue(self):
        if self.value is None:
            return None
        return ligolwtypes.ToPyType[self.type or "lstring"](self.value)


ShowTypesResultTable.RowType = ShowTypesResult



#
# =============================================================================
#
#                          Methods that implement major modes
#
# =============================================================================
#
def run_show_types(doc, connection, engine, gps_start_time, gps_end_time, included_segments_string, excluded_segments_string):
    resulttable = lsctables.New(ShowTypesResultTable)
    doc.childNodes[0].appendChild(resulttable)
    
    sql = """SELECT segment_definer.ifos, segment_definer.name, segment_definer.version,
                 (CASE WHEN segment_definer.comment IS NULL THEN '-' WHEN segment_definer.comment IS NOT NULL THEN segment_definer.comment END),
                 segment_summary.start_time, segment_summary.end_time,
                 (CASE WHEN segment_summary.comment IS NULL THEN '-' WHEN segment_summary.comment IS NOT NULL THEN segment_summary.comment END)
          FROM  segment_definer, segment_summary
          WHERE segment_definer.segment_def_id = segment_summary.segment_def_id
          AND   NOT (segment_summary.start_time > %d OR %d > segment_summary.end_time)
          """ % (gps_end_time, gps_start_time)

    rows = engine.query(sql)

    seg_dict = {}

    for row in rows:
        ifos, name, version, segment_definer_comment, segment_summary_start_time, segment_summary_end_time, segment_summary_comment = row
        key = (ifos, name, version, segment_definer_comment, segment_summary_comment)
        if key not in seg_dict:
            seg_dict[key] = []

        seg_dict[key].append(ligo.segments.segment(segment_summary_start_time, segment_summary_end_time))

    for key, value in six.iteritems(seg_dict):
        segmentlist = ligo.segments.segmentlist(value)
        segmentlist.coalesce()

        for segment in segmentlist:
            result = ShowTypesResult()
            result.ifos, result.name, result.version, result.segment_definer_comment, result.segment_summary_comment = key
            result.segment_summary_start_time, result.segment_summary_end_time = segment
            result.ifos = result.ifos.strip()
        
            resulttable.append(result)

    engine.close()


    




def run_query_types(doc, proc_id, connection, engine, gps_start_time, gps_end_time, included_segments):
    query_segment = ligo.segments.segmentlist([ligo.segments.segment(gps_start_time, gps_end_time)])

    sql = """SELECT segment_definer.ifos, segment_definer.name,segment_definer.version,
           (CASE WHEN segment_definer.comment IS NULL THEN '-' WHEN segment_definer.comment IS NOT NULL THEN segment_definer.comment END),
           segment_summary.start_time, segment_summary.end_time,
           (CASE WHEN segment_summary.comment IS NULL THEN '-' WHEN segment_summary.comment IS NOT NULL THEN segment_summary.comment END)
    FROM segment_definer, segment_summary
    WHERE segment_definer.segment_def_id = segment_summary.segment_def_id
    AND NOT(%d > segment_summary.end_time OR segment_summary.start_time > %d)
    """ % (gps_start_time, gps_end_time)

    type_clauses = list(map(seg_spec_to_sql, included_segments.split(',')))

    if type_clauses != []:
        sql += " AND (" + "OR ".join(type_clauses) + ")"


    segment_types = {}

    for row in engine.query(sql):
        sd_ifo, sd_name, sd_vers, sd_comment, ss_start, ss_end, ss_comment = row
        key = (sd_ifo, sd_name, sd_vers, sd_comment, ss_comment)
        if key not in segment_types:
            segment_types[key] = ligo.segments.segmentlist([])
        segment_types[key] |= ligo.segments.segmentlist([ligo.segments.segment(ss_start, ss_end)])

    engine.close()

    # Create segment definer and segment_summary tables
    seg_def_table = lsctables.New(lsctables.SegmentDefTable, columns = ["process_id", "segment_def_id", "ifos", "name", "version", "comment"])
    doc.childNodes[0].appendChild(seg_def_table)

    seg_sum_table = lsctables.New(lsctables.SegmentSumTable, columns = ["process_id", "segment_sum_id", "start_time", "start_time_ns", "end_time", "end_time_ns", "comment", "segment_def_id"])

    doc.childNodes[0].appendChild(seg_sum_table)

    for key in segment_types:
        # Make sure the intervals fall within the query window and coalesce
        segment_types[key].coalesce()
        segment_types[key] &= query_segment

        seg_def_id                     = seg_def_table.get_next_id()
        segment_definer                = lsctables.SegmentDef()
        segment_definer.process_id     = proc_id
        segment_definer.segment_def_id = seg_def_id
        segment_definer.ifos           = key[0]
        segment_definer.name           = key[1]
        segment_definer.version        = key[2]
        segment_definer.comment        = key[3]

        seg_def_table.append(segment_definer)

        # add each segment summary to the segment_summary_table

        for seg in segment_types[key]:
            segment_sum            = lsctables.SegmentSum()
            segment_sum.comment    = key[4]
            segment_sum.process_id = proc_id
            segment_sum.segment_def_id = seg_def_id
            segment_sum.segment_sum_id = seg_sum_table.get_next_id()
            segment_sum.start_time = seg[0]
            segment_sum.start_time_ns = 0
            segment_sum.end_time   = seg[1]
            segment_sum.end_time_ns = 0

            seg_sum_table.append(segment_sum)



def run_query_segments(doc, process_id, engine, gps_start_time, gps_end_time, include_segments, exclude_segments, result_name):
    segdefs = []

    for included in include_segments.split(','):
        spec = included.split(':')

        if len(spec) < 2 or len(spec) > 3:
            print("Included segements must be of the form ifo:name:version or ifo:name:*", file=sys.stderr)
            sys.exit(1)

        ifo     = spec[0]
        name    = spec[1]
        if len(spec) is 3 and spec[2] is not '*':
            version = int(spec[2])
            if version < 1:
                print("Segment version numbers must be greater than zero", file=sys.stderr)
                sys.exit(1)
        else:
            version = '*'

        segdefs += segmentdb_utils.expand_version_number(engine, (ifo, name, version, gps_start_time, gps_end_time, 0, 0) )

    found_segments = segmentdb_utils.query_segments(engine, 'segment', segdefs)
    found_segments = reduce(operator.or_, found_segments).coalesce()

    # We could also do:
    segment_summaries = segmentdb_utils.query_segments(engine, 'segment_summary', segdefs)

    # And we could write out everything we found
    segmentdb_utils.add_segment_info(doc, process_id, segdefs, None, segment_summaries)


    # Do the same for excluded
    if exclude_segments:
        ex_segdefs = []

        for excluded in exclude_segments.split(','):
            spec = excluded.split(':')

            if len(spec) < 2:
                print("Excluded segements must be of the form ifo:name:version or ifo:name:*", file=sys.stderr)
                sys.exit(1)

            ifo     = spec[0]
            name    = spec[1]
            version = len(spec) > 2 and spec[2] or '*'

            ex_segdefs += segmentdb_utils.expand_version_number(engine, (ifo, name, version, gps_start_time, gps_end_time, 0, 0) )


        excluded_segments = segmentdb_utils.query_segments(engine, 'segment', ex_segdefs)
        excluded_segments = reduce(operator.or_, excluded_segments).coalesce()

        found_segments.coalesce()
        found_segments -= excluded_segments



    # Add the result type to the segment definer table
    seg_name   = result_name
    seg_def_id = segmentdb_utils.add_to_segment_definer(doc, process_id, ifo, seg_name, 1)

    # and segment summary
    segmentdb_utils.add_to_segment_summary(doc, process_id, seg_def_id, [[gps_start_time, gps_end_time]])

    # and store the segments
    segmentdb_utils.add_to_segment(doc, process_id, seg_def_id, found_segments)
           


#
# =============================================================================
#
#                                 XML/File routines
#
# =============================================================================
#
class ContentHandler(ligolw.LIGOLWContentHandler):
    connection = None

def setup_files(dir_name, gps_start_time, gps_end_time):
    # Filter out the ones that are outside our time range
    xml_files = segmentdb_utils.get_all_files_in_range(dir_name, gps_start_time, gps_end_time)

    handle, temp_db  = tempfile.mkstemp(suffix='.sqlite')
    os.close(handle)

    target     = dbtables.get_connection_filename(temp_db, None, True, False)
    connection = ligolw_sqlite.setup(target)

    ContentHandler.connection = connection
    dbtables.use_in(ContentHandler)
    # With ER6 Glue change:
    ligolw_sqlite.insert_from_urls(xml_files, ContentHandler)

    # Pre-ER6 Kipp Glue change:
    #ligolw_sqlite.insert_from_urls(connection, xml_files) # [temp_xml])

    segmentdb_utils.ensure_segment_table(connection)
        
    return temp_db, connection
    



#
# =============================================================================
#
#                                     Main
#
# =============================================================================
#

if __name__ == '__main__':
    options, database_location, file_location  = parse_command_line()    

    # Ping the database and exit if requested
    if options.ping:
        connection = segmentdb_utils.setup_database(database_location)
        print(connection.ping())
        sys.exit( 0 )

    gps_start_time = int(options.gps_start_time)
    gps_end_time   = int(options.gps_end_time)

    # set up the response
    doc = ligolw.Document()
    doc.appendChild(ligolw.LIGO_LW())
    process_id = process.register_to_xmldoc(doc, PROGRAM_NAME, options.__dict__, version = git_version.id, cvs_entry_time = __date__).process_id

    temp_files = False

    if database_location:
        connection = segmentdb_utils.setup_database(database_location)
        engine     = query_engine.LdbdQueryEngine(connection)
    else:
        temp_db, connection = setup_files(file_location, gps_start_time, gps_end_time)
        engine     = query_engine.SqliteQueryEngine(connection)
        temp_files = True

    
    if options.show_types:
        run_show_types(doc, connection, engine, gps_start_time, gps_end_time,
                       options.include_segments,options.exclude_segments)
        

    if options.query_segments:
        run_query_segments(doc, process_id, engine, gps_start_time, gps_end_time,
                           options.include_segments, options.exclude_segments,
                           options.result_name)


    if options.query_types:
        run_query_types(doc, process_id, connection, engine, gps_start_time, gps_end_time, options.include_segments)


    gz = (options.output_file is not None and
          options.output_file.endswith(".gz"))
    utils.write_filename(doc, options.output_file, gz=gz)

    # Clean up
    if temp_files:
        os.remove(temp_db)





"""
ldbdd listens on a socket for connections from a client, processes the
requests and then returns the results to the client.

The server is built on top of the ThreadingGSITCPSocketServer from
the io module in U{pyGlobus<http://www-itg.lbl.gov/gtg/projects/pyGlobus/>},
which is build on top of the standard module SocketServer.

This program is part of the Grid LSC User Environment (GLUE)

GLUE is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import print_function
import os
import sys
import signal
import getopt
import time
import socket
import six.moves.configparser
import logging
import logging.handlers
from pyGlobus import io
from glue import gsiserverutils


def SIGHUPhandler(signum, frame):
  """
  Handle SIGHUP signals by asking the server to stop its 
  service, then read the gridmap file again and reconfigure, 
  after which the service will automatically start again.

  @param signum: signal number; see the documentation for the 
    signal module

  @param frame: current stack frame; see the documentation for
    the signal module
        
  @return: None
  """
  global logger
  global myServer
  global myConfigParser
  
  myServer.stopService()
  logger.info("Re-reading configuration file")

  myConfigParser.read(configFilePath)
  for k in configuration.keys():
    try:
      value = myConfigParser.get('ldbdd',k)
    except six.moves.configparser.NoOptionError:
      try:
        logger.error("Error: missing configuration option : %s" % (k))
      except:
        print("Error: missing configuration option : %s" % (k), file=sys.stderr)
      sys.exit(1)
    try:
      configuration[k] = eval(value)
    except:
      configuration[k] = value

  logger.info("Re-initializing server module")
  ServerModule.shutdown()
  myServer.initialize()
  ServerModule.initialize(configuration,logger)

def SIGTERMhandler(signum, frame):
  """
  Handle SIGTERM signals by asking the server to stop and die.

  @param signum: signal number; see the documentation for the 
    signal module

  @param frame: current stack frame; see the documentation for
    the signal module
        
  @return: None
  """
  global myServer

  myServer.die()


def socketReadyCallback(arg, handle, result):
  """
  Called when a socket is ready for reading (when a connection has been made
  and there is something to read).

  This function simply sets a flag to let the server loop know the socket
  is ready for reading.

  @param arg: User supplied argument. Here it is used to carry the socket ready
    flag.
  @param handle: a pointer to the SWIG'ized globus_io_handle_t
  @param result: a pointer to the SWIG'ized globus_result_t

  @return: None
  """
  arg.mySocketReady = 1


# parse command line options
shortop = "c:dh"
longop = [
  "config-file=",
  "daemon",
  "help"
  ]


usage = """\
Usage: ldbdd [OPTIONS]

  -c, --config-file FILE    read configuration from FILE
  -d, --daemon              run as a deamon
  -h, --help                print detailed help message

If no configuration file is given, the server will attempt to use

  ${GLUE_PREFIX}/etc/ldbdd.ini

Sending the server process a HUP will cause the confuration and grid-mapfile
to be re-read and sending the process a TERM will shutdown the server.
"""

#default for command-line options
runAsDaemon = False
configFilePath = None

try:
  opts, args = getopt.getopt(sys.argv[1:], shortop, longop)
except getopt.GetoptError:
  print("Error parsing command line", file=sys.stderr)
  sys.exit(1)

for o, a in opts:
  if o in ("-d", "--daemon"):
    runAsDaemon = True
  if o in ("-c", "--config-file"):
    configFilePath = a
  if o in ("-h", "--help"):
    print(usage)
    sys.exit(0)

try:
  GLUE_PREFIX = os.environ["GLUE_PREFIX"]
except:
  pass

if not configFilePath:
  configFilePath = os.path.join(GLUE_PREFIX, "etc/ldbdserver.ini")

# default configuration values may go here, but they will be replaced
# by actual values if and when a ConfigurationManager instance is
# created elsewhere
configuration = {
  'server' : None,
  'ipaddr' : '',
  'port' : 30015,
  'secure' : 'yes',
  'gridmap' : 'grid-mapfile',
  'certfile' : 'ldbdcert.pem',
  'keyfile' : 'ldbdkey.pem',
  'dbname' : 'ldbd_tst',
  'dbuser' : 'grid',
  'rls' : 'rls://kitalpha.ligo.caltech.edu',
  'dbpasswd' : '',
  'max_client_byte_string': 1048576,
  'pidfile' : 'ldbdd.pid',
  'logfile' : 'ldbdd.log',
  'logmaxbytes' : 1024 * 1024 * 1,
  'logbackupcount' : 5,
  'loglevel' : 'INFO',
  'ldbd_com':'',
  'db2_com':''
  }


# grab configuration
myConfigParser = six.moves.configparser.ConfigParser()
try:
  myConfigParser.read(configFilePath)
except:
  print("Error: unable to read configuration file : %s", file=sys.stderr)
  sys.exit(1)

for k in configuration.keys():
  try:
    value = myConfigParser.get('ldbdd',k)
  except six.moves.configparser.NoOptionError:
    print("Error: missing configuration option : %s" % (k), file=sys.stderr)
    sys.exit(1)
  try:
    configuration[k] = eval(value)
  except:
    configuration[k] = value

# setup to cach HUP and SIGTERM
signal.signal(signal.SIGHUP, SIGHUPhandler)
signal.signal(signal.SIGTERM, SIGTERMhandler)

# become a daemon
if runAsDaemon:
  gsiserverutils.daemon()

# set up logging but don't set handler here since that is
# configurable in the .ini file
logger = logging.getLogger('ldbdd')


class LDBDDaemonException(Exception):
  """
  Class representing exceptions withing the LDBDDaemon class.
  """
  def __init__(self, args=None):
    """
    Initialize an instance.

    @param args: 

    @return: Instance of class ServerHandlerException
    """
    self.args = args


class LDBDDaemon(object):
  """
  An instance of this class is a multi-threadead server that will listen on a
  socket/port for connections from a client and will process requests.
  """
  def __init__(self,handler):
    """
    Perform any checks that are necessary before becoming a daemon and
    starting a socket service to listen on. If any check fail print
    to stderr and exit.

    @param:

    @return: instance of class LDBDDaemon
    """
    self.serverName = None
    self.handler = handler
    self.server = None

    # verify that we have access to the certificate, key, grid-mapfile
    certFilePath = configuration["certfile"]
    if not os.access(certFilePath, os.R_OK):
      print("ldbdd: " + \
      "Cannot access certificate file at %s" % certFilePath, file=sys.stderr)
      sys.exit(1)

    keyFilePath = configuration["keyfile"]
    if not os.access(keyFilePath, os.R_OK):
      print("ldbdd: Cannot access key file at %s" % keyFilePath, file=sys.stderr)
      sys.exit(1)

    gridmapFilePath = configuration["gridmap"]
    if not os.access(gridmapFilePath, os.R_OK):
      print("ldbdd: Cannot access grid-mapfile at %s" % gridmapFilePath, file=sys.stderr)
      sys.exit(1)

    # verify that we can write PID file
    failure = 0
    pidFilePath = configuration["pidfile"]
    exists  = os.access(pidFilePath, os.F_OK)
    if exists:
      if not os.access(pidFilePath, os.W_OK):
        failure = 1
    else:
      try:
        f = open(pidFilePath, "w")
        f.close()
      except:
        failure = 1
                        
    if failure:
      print("ldbdd: Cannot write PID to file %s" % pidFilePath, file=sys.stderr)
      sys.exit(1)

  def initialize(self):
    """
    Grab information from configuration and record it for this instance. Set
    up logging.

    @param:
                
    @return: None

    """
    global logger

    if self.serverName is None:
      msg = "Error: the server name is not configured"
      raise LDBDDaemonException(msg)
    if self.serverName != configuration["server"]:
      msg = "Error: attempt to change service %s into service %s" % \
        ( self.serverName, configuration["server"] )
      raise LDBDDaemonException(msg)

    try:
      self.certFilePath = configuration["certfile"]
      self.keyFilePath  = configuration["keyfile"]
      self.gridmapFilePath = configuration["gridmap"]
      self.pidFilePath = configuration["pidfile"]
      self.ipaddr = configuration["ipaddr"]
      self.port = configuration["port"]
      self.secure = configuration["secure"]

      myLogger = logging.getLogger('ldbdd')

      # remove any existing handlers
      for h in myLogger.handlers:
        myLogger.removeHandler(h)

      logFilePath = configuration["logfile"]

      handler = logging.handlers.RotatingFileHandler(
        logFilePath, 'a', configuration['logmaxbytes'], 
        configuration['logbackupcount'])
      formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
      handler.setFormatter(formatter)
      myLogger.addHandler(handler)
      logFileLevel = configuration["loglevel"]
      myLogger.setLevel(eval("logging." + logFileLevel))

      self.logger = myLogger
      logger = myLogger

    except Exception as e:
      print("Unable to start logging: %s" % e, file=sys.__stderr__)
        
    for k in configuration.keys():
      logger.info("Configuration parameter: %s = %s" %
        (k, str(configuration[k])))

  def run(self):
    """
    Read configuration information in the initialize() method, write out
    pid, set up our environment, start the GSI TCP service on the socket,
    then start listening forever.

    @param:

    @return: None
    """
    self.writePID()
    self.setEnvironment()
    self.startService()
    self.serveForever()

  def stopService(self):
    """
    Set a flag on this instance of LDBDDaemon so that when the main loop in
    the serveForever() method sees the flag set the service is stopped.

    @param:

    @return: None
    """
    self.logger.info("stopNow flag set; server will stop during next loop")
    self.stopNow = 1
        
  def die(self):
    """
    Set a flag on this instance of LDBDDaemon so that when the main loop in
    the serveForever() method sees the flag set the service is stopped and
    then the executable exits.

    @param:

    @return: None
    """
    self.logger.info("dieNow flag set; executable will exit during next loop")
    self.stopService()
    self.dieNow = 1

  def writePID(self):
    """
    Write pid out to a file.
                
    @param:

    @return: None
    """

    # write PID out to file
    myPIDfd = os.open(self.pidFilePath, os.O_WRONLY | os.O_CREAT, 0o644)
    os.write(myPIDfd, "%d\n" % os.getpid())
    os.close(myPIDfd)

  def clearPID(self):
    """
    Delete pid file.
                
    @param:

    @return: None
    """

    try:
      os.unlink(self.pidFilePath)
    except:
      pass

  def setEnvironment(self):
    """
    Set environment variables that this instance of ldbdd needs, in particular
    for GSI services.
                
    @param:

    @return: None
    """
    os.environ["X509_USER_CERT"] = self.certFilePath
    os.environ["X509_USER_KEY"] = self.keyFilePath
    os.environ["GRIDMAP"] = self.gridmapFilePath

  def startService(self):
    """
    Create instance of io.ThreadingGSITCPSocketServer that this server will
    use to listen for incoming requests from clients. See the
    U{pyGlobus<http://www-itg.lbl.gov/gtg/projects/pyGlobus/>} documentation.

    If the ThreadingGSITCPSocketServer cannot be started sleep and then try
    again, with the time between retries growing exponentially.

    This server requires GSI authentication. 

    @param:

    @return: None
    """
    # create TCPIO Attribute object and initialize to pass to server        
    tcpIOAttr = io.TCPIOAttr()
    authData = io.AuthData()

    # create a callback to use for authorization
    gridmap = gsiserverutils.Gridmap(self.gridmapFilePath, self.logger)
    callback = gsiserverutils.AuthCallback(gridmap, self.logger, callback=None)

    # set authorization callback
    authData.set_callback(callback, self)
    self.authData = authData

    if self.secure == 'no':
      # set up an unathenticated connection
      self.logger.info("Disabling security")
      tcpIOAttr.set_authentication_mode(
        io.ioc.GLOBUS_IO_SECURE_AUTHENTICATION_MODE_NONE)
      tcpIOAttr.set_authorization_mode(
        io.ioc.GLOBUS_IO_SECURE_AUTHORIZATION_MODE_NONE, authData)
      tcpIOAttr.set_channel_mode(
        io.ioc.GLOBUS_IO_SECURE_CHANNEL_MODE_CLEAR)
      tcpIOAttr.set_delegation_mode(
        io.ioc.GLOBUS_IO_SECURE_DELEGATION_MODE_NONE)

    else:
      # we use secure IO authentication using GSSAPI and authorization back
      # to a callback function, which reads a gridmap file
      self.logger.info("Enabling security")
      tcpIOAttr.set_authentication_mode(
        io.ioc.GLOBUS_IO_SECURE_AUTHENTICATION_MODE_GSSAPI)
      tcpIOAttr.set_authorization_mode(
        io.ioc.GLOBUS_IO_SECURE_AUTHORIZATION_MODE_CALLBACK, authData)
      tcpIOAttr.set_channel_mode(
        io.ioc.GLOBUS_IO_SECURE_CHANNEL_MODE_CLEAR)
      tcpIOAttr.set_delegation_mode(
        io.ioc.GLOBUS_IO_SECURE_DELEGATION_MODE_FULL_PROXY)

    self.tcpIOAttr = tcpIOAttr

    # bind to an interface, if specified
    if self.ipaddr is not '':
      self.logger.info("Listening on ip address: %s" % self.ipaddr)
      self.tcpIOAttr.set_interface(self.ipaddr)
    else:
      self.logger.info("Listening on all interfaces")

    # start the server
    self.running = 0
    self.nap = 1
    while not self.running:
      try:
        self.logger.info("Starting server on port %d" % self.port)
        server = io.ThreadingGSITCPSocketServer(
          addr=(self.ipaddr, self.port), 
          RequestHandlerClass=self.handler,
          channel_mode=io.ioc.GLOBUS_IO_SECURE_CHANNEL_MODE_CLEAR,
          delegation_mode=io.ioc.GLOBUS_IO_SECURE_DELEGATION_MODE_FULL_PROXY,
          tcpAttr=self.tcpIOAttr
          )
        self.server = server
        self.logger.info("Server running on port %d" % self.port)

        self.running = 1
        self.nap = 1
                                
        self.stopNow = 0
        self.dieNow = 0
                        
      except Exception as e:
        self.logger.warning("Error starting server: %s" % e)
        self.logger.warning("Will retry in %d seconds" % self.nap)

        try:
          del server
          del self.server
        except:
          pass

        time.sleep(self.nap)
        self.nap = self.nap * 2
                        
  def serveForever(self):
    """
    Our own version of the server_forever() method for the GSITCPServer and
    SocketServer.BaseServer classes. Normally the call sequence is

    get_request()
    verify_request()
    process_request()

    and this is normally done with error handling by the handle_request() 
    method. In turn server_forever() is usually just

    while 1: handle_request()

    The get_request() for GSITCPServer is normally a blocking listen() on the
    socket followed by the accept() call.

    In order to not block we instead use a register_listen() call and have the
    callback set a flag when a socket is ready to be answered and a call to
    accept() made.

    Since the loop is not blocking on the socket IO, it can be interrupted
    by signals, such as a SIGHUP.

    @param:
        
    @return: None
    """
    # set socket ready flag to false and register a listening callback
    # that is called when the socket is ready
    self.mySocketReady = 0
    handle = self.server.socket.register_listen(socketReadyCallback, self)

    while self.running:
      # is my socket ready? 
      if self.mySocketReady:
        # socket is ready so accept the connection, then process it
        # the process_request() method used here will be that from
        # the SocketServer.ThreadingMixIn class so inside of that a
        # new thread is started.
        try:
          self.logger.debug("calling accept method for socket instance")
          (request, client_address) = self.server.socket.accept(
            self.server.attr)
          process = 1
        except io.GSITCPSocketException as ex:
          self.server.socket.shutdown(2)
          process = 0

        if process:
          try:
            self.logger.debug("processing request on socket now...")
            self.server.process_request(request, client_address)
            self.logger.debug("finished processing request")
          except Exception as e:
            self.logger.error("Error during server.process_request(): %s" % e)
            self.server.handle_error(request, client_address)
            self.server.close_request(request)

        # free callback handle used for the register_listen
        self.server.socket.free_callback(handle)

        # prepare for next connection by setting ready flag to false and
        # registering a new listener
        self.mySocketReady = 0
        handle = self.server.socket.register_listen(socketReadyCallback, self)
        continue

      # have I been told to shutdown?
      if self.stopNow:
        # close and delete the server
        self.logger.info('closing server')
        self.server.server_close()
        del self.server

        # have I been told to die?
        if self.dieNow:
          self.clearPID()
          raise SystemExit

        # give 5 seconds for the socket to be freed up
        time.sleep(5)
        self.running = 0

        continue

      # no socket is ready and I have not been told to shutdown
      # so sleep for a bit so that we don't burn CPU all the time
      time.sleep(0.1)


# figure out what server we are using and load it
serv = configuration['server']
ServerModule = __import__( "glue." + serv, globals(), locals(), [serv] )

# initialize daemon inheriting the server properties from the loaded module
myServer = LDBDDaemon(ServerModule.ServerHandler)
myServer.serverName = configuration['server']
myServer.initialize()
ServerModule.initialize(configuration,logger)

running = 1

try:
  while running:
    # start server
    myServer.run()

except SystemExit:
  ServerModule.shutdown()
  logger.info("ldbdd shutting down")
  sys.exit()

except KeyboardInterrupt:
  ServerModule.shutdown()
  logger.info("ldbdd shutting down")
  sys.exit()

except Exception as e:
  try:
    ServerModule.shutdown()
  except:
    pass
  msg = "ldbdd is stopping due to unhandled error: %s" % e
  print(msg, file=sys.stderr)
  logger.critical(msg)

del logger
logger = None
sys.exit()
